# Copyright (c) 2024-2025, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from __future__ import annotations

from ... import _bindings
from ... import _cccl_interop as cccl
from ..._caching import cache_with_registered_key_functions
from ..._cccl_interop import call_build, set_cccl_iterator_state
from ..._utils.protocols import (
    get_data_pointer,
    get_dtype,
    validate_and_get_stream,
)
from ..._utils.temp_storage_buffer import TempStorageBuffer
from ...typing import DeviceArrayLike
from ._sort_common import DoubleBuffer, SortOrder, _get_arrays


class _RadixSort:
    __slots__ = [
        "d_in_keys_cccl",
        "d_out_keys_cccl",
        "d_in_values_cccl",
        "d_out_values_cccl",
        "decomposer_op",
        "build_result",
    ]

    def __init__(
        self,
        d_in_keys: DeviceArrayLike | DoubleBuffer,
        d_out_keys: DeviceArrayLike | None,
        d_in_values: DeviceArrayLike | DoubleBuffer | None,
        d_out_values: DeviceArrayLike | None,
        order: SortOrder,
    ):
        d_in_keys_array, d_out_keys_array, d_in_values_array, d_out_values_array = (
            _get_arrays(d_in_keys, d_out_keys, d_in_values, d_out_values)
        )

        self.d_in_keys_cccl = cccl.to_cccl_input_iter(d_in_keys_array)
        self.d_out_keys_cccl = cccl.to_cccl_output_iter(d_out_keys_array)
        self.d_in_values_cccl = cccl.to_cccl_input_iter(d_in_values_array)
        self.d_out_values_cccl = cccl.to_cccl_output_iter(d_out_values_array)

        # TODO: decomposer op is not supported for now
        self.decomposer_op = cccl.Op(
            name="",
            operator_type=cccl.OpKind.STATELESS,
            ltoir=b"",
            state_alignment=1,
            state=None,
        )
        decomposer_return_type = "".encode("utf-8")

        self.build_result = call_build(
            _bindings.DeviceRadixSortBuildResult,
            _bindings.SortOrder.ASCENDING
            if order is SortOrder.ASCENDING
            else _bindings.SortOrder.DESCENDING,
            self.d_in_keys_cccl,
            self.d_in_values_cccl,
            self.decomposer_op,
            decomposer_return_type,
        )

    def __call__(
        self,
        temp_storage,
        d_in_keys: DeviceArrayLike | DoubleBuffer,
        d_out_keys: DeviceArrayLike | None,
        d_in_values: DeviceArrayLike | DoubleBuffer | None,
        d_out_values: DeviceArrayLike | None,
        num_items: int,
        begin_bit: int | None = None,
        end_bit: int | None = None,
        stream=None,
    ):
        d_in_keys_array, d_out_keys_array, d_in_values_array, d_out_values_array = (
            _get_arrays(d_in_keys, d_out_keys, d_in_values, d_out_values)
        )

        set_cccl_iterator_state(self.d_in_keys_cccl, d_in_keys_array)
        if d_in_values_array is not None:
            set_cccl_iterator_state(self.d_in_values_cccl, d_in_values_array)
        set_cccl_iterator_state(self.d_out_keys_cccl, d_out_keys_array)
        if d_out_values_array is not None:
            set_cccl_iterator_state(self.d_out_values_cccl, d_out_values_array)

        is_overwrite_okay = isinstance(d_in_keys, DoubleBuffer)

        stream_handle = validate_and_get_stream(stream)
        if temp_storage is None:
            temp_storage_bytes = 0
            d_temp_storage = 0
        else:
            temp_storage_bytes = temp_storage.nbytes
            # Note: this is slightly slower, but supports all ndarray-like objects as long as they support CAI
            # TODO: switch to use gpumemoryview once it's ready
            d_temp_storage = get_data_pointer(temp_storage)

        if begin_bit is None:
            begin_bit = 0
        if end_bit is None:
            key_type = get_dtype(d_in_keys_array)
            end_bit = key_type.itemsize * 8

        selector = -1

        temp_storage_bytes, selector = self.build_result.compute(
            d_temp_storage,
            temp_storage_bytes,
            self.d_in_keys_cccl,
            self.d_out_keys_cccl,
            self.d_in_values_cccl,
            self.d_out_values_cccl,
            self.decomposer_op,
            num_items,
            begin_bit,
            end_bit,
            is_overwrite_okay,
            selector,
            stream_handle,
        )

        if is_overwrite_okay and temp_storage is not None:
            assert selector in (0, 1)
            assert isinstance(d_in_keys, DoubleBuffer)
            d_in_keys.selector = selector
            if d_in_values is not None:
                assert isinstance(d_in_values, DoubleBuffer)
                d_in_values.selector = selector

        return temp_storage_bytes


@cache_with_registered_key_functions
def make_radix_sort(
    d_in_keys: DeviceArrayLike | DoubleBuffer,
    d_out_keys: DeviceArrayLike | None,
    d_in_values: DeviceArrayLike | DoubleBuffer | None,
    d_out_values: DeviceArrayLike | None,
    order: SortOrder,
):
    """Implements a device-wide radix sort using ``d_in_keys`` in the requested order.

    Example:
        Below, ``make_radix_sort`` is used to create a radix sort object that can be reused.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/sort/radix_sort_object.py
          :language: python
          :start-after: # example-begin


    Args:
        d_in_keys: Device array or DoubleBuffer containing the input keys to be sorted
        d_out_keys: Device array to store the sorted keys
        d_in_values: Optional Device array or DoubleBuffer containing the input keys to be sorted
        d_out_values: Device array to store the sorted values
        op: Callable representing the comparison operator

    Returns:
        A callable object that can be used to perform the radix sort
    """
    return _RadixSort(d_in_keys, d_out_keys, d_in_values, d_out_values, order)


def radix_sort(
    d_in_keys: DeviceArrayLike | DoubleBuffer,
    d_out_keys: DeviceArrayLike | None,
    d_in_values: DeviceArrayLike | DoubleBuffer | None,
    d_out_values: DeviceArrayLike | None,
    order: SortOrder,
    num_items: int,
    begin_bit: int | None = None,
    end_bit: int | None = None,
    stream=None,
):
    """
    Performs device-wide radix sort.

    This function automatically handles temporary storage allocation and execution.

    Example:
        Below, ``radix_sort`` is used to sort a sequence of keys. It also rearranges the values according to the keys' order.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/sort/radix_sort_basic.py
            :language: python
            :start-after: # example-begin


        In the following example, ``radix_sort`` is used to sort a sequence of keys with a ``DoubleBuffer`` for reduced temporary storage.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/sort/radix_sort_buffer.py
            :language: python
            :start-after: # example-begin


    Args:
        d_in_keys: Device array or DoubleBuffer containing the input sequence of keys
        d_out_keys: Device array to store the sorted keys (optional)
        d_in_values: Device array or DoubleBuffer containing the input sequence of values (optional)
        d_out_values: Device array to store the sorted values (optional)
        order: Sort order (ascending or descending)
        num_items: Number of items to sort
        begin_bit: Beginning bit position for comparison (optional)
        end_bit: Ending bit position for comparison (optional)
        stream: CUDA stream for the operation (optional)
    """
    sorter = make_radix_sort(d_in_keys, d_out_keys, d_in_values, d_out_values, order)
    tmp_storage_bytes = sorter(
        None,
        d_in_keys,
        d_out_keys,
        d_in_values,
        d_out_values,
        num_items,
        begin_bit,
        end_bit,
        stream,
    )
    tmp_storage = TempStorageBuffer(tmp_storage_bytes, stream)
    sorter(
        tmp_storage,
        d_in_keys,
        d_out_keys,
        d_in_values,
        d_out_values,
        num_items,
        begin_bit,
        end_bit,
        stream,
    )



# Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

"""CacheModifiedInputIterator implementation."""

from __future__ import annotations

import struct
from textwrap import dedent
from typing import Literal

from .._bindings import Op, OpKind
from .._cpp_compile import compile_cpp_to_ltoir, cpp_type_from_descriptor
from .._utils.protocols import get_data_pointer, get_dtype
from ..types import from_numpy_dtype
from ._base import IteratorBase
from ._common import CUDA_PREAMBLE

# Map modifier names to PTX cache operators and C++ intrinsics
_CACHE_MODIFIERS = {
    "stream": ("cs", "__ldcs"),  # Cache streaming (evict first)
    "global": ("cg", "__ldcg"),  # Cache at L2 only
    "volatile": ("cv", "__ldcv"),  # Don't cache, always fetch
}


class CacheModifiedInputIterator(IteratorBase):
    """
    Iterator that wraps a device pointer with cache-modified loads.

    This iterator uses PTX cache modifiers to control how data is loaded:
    - "stream": Uses streaming loads (ld.global.cs) - hints that data will not be reused
    - "global": Uses global cache loads (ld.global.cg) - caches only at L2
    - "volatile": Uses volatile loads (ld.global.cv) - always fetches from memory

    Supports element types of size 1, 2, 4, 8, or 16 bytes.
    """

    __slots__ = [
        "_modifier",
        "_array",
        "_ptr",
    ]

    def __init__(
        self,
        array,
        modifier: Literal["stream", "global", "volatile"] = "stream",
    ):
        """
        Create a cache-modified input iterator.

        Args:
            array: Device array to wrap (must support __cuda_array_interface__)
            modifier: Cache modifier - "stream", "global", or "volatile"
        """
        if modifier not in _CACHE_MODIFIERS:
            raise ValueError(
                f"Unknown modifier: {modifier}. Must be one of {list(_CACHE_MODIFIERS.keys())}"
            )

        self._modifier = modifier
        self._array = array  # Keep reference to prevent GC
        ptr = get_data_pointer(array)
        dtype = get_dtype(array)

        self._ptr = ptr
        value_type = from_numpy_dtype(dtype)

        # Cache-modified loads only supported for power-of-two sizes up to 16 bytes
        # These correspond to PTX instructions: ld.global.{modifier}.b{8,16,32,64,128}
        if value_type.size not in (1, 2, 4, 8, 16):
            raise ValueError(
                f"CacheModifiedInputIterator only supports types of size 1, 2, 4, 8, or 16 bytes. "
                f"Got type with size {value_type.size} bytes. "
                f"This matches PTX cache-modified load instruction limitations."
            )

        # State is just the pointer (8 bytes on 64-bit)
        state_bytes = struct.pack("Q", ptr)

        super().__init__(
            state_bytes=state_bytes,
            state_alignment=8,  # Pointer alignment
            value_type=value_type,
        )

    def _make_advance_op(self) -> Op:
        symbol = self._make_advance_symbol()
        cpp_type = cpp_type_from_descriptor(self._value_type)

        source = dedent(f"""
            {CUDA_PREAMBLE}

            extern "C" __device__ void {symbol}(void* state, void* offset) {{
                auto* s = static_cast<{cpp_type}**>(state);
                auto dist = *static_cast<uint64_t*>(offset);
                *s += dist;
            }}
        """).strip()

        ltoir = compile_cpp_to_ltoir(source)
        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[],
        )

    def _make_input_deref_op(self) -> Op | None:
        symbol = self._make_input_deref_symbol()
        cpp_type = cpp_type_from_descriptor(self._value_type)
        _, intrinsic = _CACHE_MODIFIERS[self._modifier]

        # Use cache-modified intrinsic for all supported sizes (1, 2, 4, 8, 16 bytes)
        # These correspond to PTX instructions: ld.global.{modifier}.b{8,16,32,64,128}
        # Note: __ldcs, __ldcg, __ldcv intrinsics work for all these sizes
        source = dedent(f"""
            {CUDA_PREAMBLE}

            extern "C" __device__ void {symbol}(void* state, void* result) {{
                auto* ptr = *static_cast<{cpp_type}**>(state);
                *static_cast<{cpp_type}*>(result) = {intrinsic}(ptr);
            }}
        """).strip()

        ltoir = compile_cpp_to_ltoir(source)
        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[],
        )

    def _make_output_deref_op(self) -> Op | None:
        # Cache-modified iterator is input-only
        return None

    def __add__(self, offset: int) -> "CacheModifiedInputIterator":
        """Advance the iterator by offset elements."""
        out = CacheModifiedInputIterator(self._array, self._modifier)
        offset_ptr = self._ptr + offset * get_dtype(out._array).itemsize
        out._ptr = offset_ptr
        out._state_bytes = struct.pack("Q", offset_ptr)
        out._uid_cached = None
        return out

    @property
    def kind(self):
        """Return a hashable kind for caching purposes."""
        return (
            "CacheModifiedInputIterator",
            self._modifier,
            self._value_type,
        )



# Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from __future__ import annotations

from ._base import IteratorBase
from ._cache_modified import CacheModifiedInputIterator
from ._constant import ConstantIterator
from ._counting import CountingIterator
from ._discard import DiscardIterator
from ._permutation import PermutationIterator
from ._reverse import ReverseIterator
from ._shuffle import ShuffleIterator
from ._transform import TransformIterator, TransformOutputIterator
from ._zip import ZipIterator

__all__ = [
    "CacheModifiedInputIterator",
    "ConstantIterator",
    "CountingIterator",
    "DiscardIterator",
    "IteratorBase",
    "PermutationIterator",
    "ReverseIterator",
    "ShuffleIterator",
    "TransformIterator",
    "TransformOutputIterator",
    "ZipIterator",
]



# Copyright (c) 2024-2025, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

"""
cuda.coop.block_scan
===========================

This module provides a set of :ref:`collective <collective-primitives>`
for computing parallel prefix scans of items partitioned across CUDA
thread blocks.  It is based on the :class:`cub.BlockScan` C++ class in
the CUB library.

Supported C++ APIs
++++++++++++++++++

The following :class:`cub.BlockScan` C++ APIs are supported:


    ExclusiveSum(T input, T &output)
    ExclusiveSum(T input, T &output, BlockPrefixCallbackOp &prefix_op)
    ExclusiveSum(T&)[ITEMS_PER_THREAD] input, T(&)[ITEMS_PER_THREAD] output)
    ExclusiveSum(T&)[ITEMS_PER_THREAD] input, T(&)[ITEMS_PER_THREAD] output, BlockPrefixCallbackOp &prefix_op)

    InclusiveSum(T input, T &output)
    InclusiveSum(T input, T &output, BlockPrefixCallbackOp &prefix_op)
    InclusiveSum(T&)[ITEMS_PER_THREAD] input, T(&)[ITEMS_PER_THREAD] output)
    InclusiveSum(T&)[ITEMS_PER_THREAD] input, T(&)[ITEMS_PER_THREAD] output, BlockPrefixCallbackOp &prefix_op)

    ExclusiveScan(T input, T &output, ScanOp scan_op)
    ExclusiveScan(T input, T &output, ScanOp scan_op, BlockPrefixCallbackOp &prefix_op)
    ExclusiveScan(T input, T &output, T initial_value, ScanOp scan_op)
    ExclusiveScan(T&)[ITEMS_PER_THREAD] input, T(&)[ITEMS_PER_THREAD] output, ScanOp scan_op)
    ExclusiveScan(T&)[ITEMS_PER_THREAD] input, T(&)[ITEMS_PER_THREAD] output, ScanOp scan_op, BlockPrefixCallbackOp &prefix_op)
    ExclusiveScanT(&)[ITEMS_PER_THREAD] input, T(&)[ITEMS_PER_THREAD] output, T initial_value, ScanOp scan_op)

    InclusiveScan(T input, T &output, ScanOp scan_op)
    InclusiveScan(T input, T &output, ScanOp scan_op, BlockPrefixCallbackOp &prefix_op)
    InclusiveScan(T&)[ITEMS_PER_THREAD] input, T(&)[ITEMS_PER_THREAD] output, ScanOp scan_op)
    InclusiveScan(T&)[ITEMS_PER_THREAD] input, T(&)[ITEMS_PER_THREAD] output, ScanOp scan_op, BlockPrefixCallbackOp &prefix_op)
    InclusiveScanT(&)[ITEMS_PER_THREAD] input, T(&)[ITEMS_PER_THREAD] output, T initial_value, ScanOp scan_op)

Unsupported C++ APIs
++++++++++++++++++++

This module does not support any of the :class:`cub.BlockScan` C++ APIs
that take a block aggregate reference as an argument.  That being said, the
`BlockPrefixCallbackOp` callable is supported, and thus, block aggregates can
be obtained using those measures.

The reason the `T &block_aggregate` pattern is not supported as it will usually
result in two output parameters, which we don't support in our underlying type
machinery (i.e. _types.py).

The unsupported APIs are as follows:

    ExclusiveSum(T input, T &output, T &block_aggregate)
    ExclusiveSum(T&)[ITEMS_PER_THREAD] input, T(&)[ITEMS_PER_THREAD] output, T &block_aggregate)

    InclusiveSum(T input, T &output, T &block_aggregate)
    InclusiveSum(T&)[ITEMS_PER_THREAD] input, T(&)[ITEMS_PER_THREAD] output, T &block_aggregate)

    ExclusiveScan(T input, T &output, ScanOp scan_op, T &block_aggregate)
    ExclusiveScan(T input, T &output, T initial_value, ScanOp scan_op, T &block_aggregate)
    ExclusiveScan(T&)[ITEMS_PER_THREAD] input, T(&)[ITEMS_PER_THREAD] output, ScanOp scan_op, T &block_aggregate)
    ExclusiveScan(T&)[ITEMS_PER_THREAD] input, T(&)[ITEMS_PER_THREAD] output, T initial_value, ScanOp scan_op, T &block_aggregate)

    InclusiveScan(T input, T &output, ScanOp scan_op, T &block_aggregate)
    InclusiveScan(T&)[ITEMS_PER_THREAD] input, T(&)[ITEMS_PER_THREAD] output, T initial_value, ScanOp scan_op, T &block_aggregate)
"""

from typing import Any, Callable, Literal

import numba

from .._common import (
    CUB_BLOCK_SCAN_ALGOS,
    make_binary_tempfile,
    normalize_dim_param,
    normalize_dtype_param,
)
from .._scan_op import (
    ScanOp,
)
from .._types import (
    Algorithm,
    Dependency,
    DependentArray,
    DependentCxxOperator,
    DependentPythonOperator,
    DependentReference,
    Invocable,
    Pointer,
    TemplateParameter,
    numba_type_to_wrapper,
)
from .._typing import (
    DimType,
    DtypeType,
    ScanOpType,
)


def make_scan(
    dtype: DtypeType,
    threads_per_block: DimType,
    items_per_thread: int,
    initial_value: Any = None,
    mode: Literal["exclusive", "inclusive"] = "exclusive",
    scan_op: ScanOpType = "+",
    block_prefix_callback_op: Callable = None,
    algorithm: Literal["raking", "raking_memoize", "warp_scans"] = "raking",
    methods: dict = None,
) -> Callable:
    """
    Creates a block-wide prefix scan primitive based on CUB's BlockScan
    APIs.

    The returned primitive is callable from a Numba CUDA kernel and
    supports sum and generic scan operators in inclusive and exclusive
    modes.

    Example:
        The snippet below creates a scan primitive and invokes the
        returned ``block_scan`` primitive inside a kernel.

        .. code-block:: python

           block_scan = coop.block.make_scan(
               dtype=numba.int32,
               threads_per_block=128,
               items_per_thread=4,
               mode="exclusive",
               scan_op="+",
           )

           @cuda.jit(link=block_scan.files)
           def kernel(input_arr, output_arr):
               tid = cuda.threadIdx.x
               thread_data = cuda.local.array(4, dtype=numba.int32)
               for i in range(4):
                   thread_data[i] = input_arr[tid * 4 + i]
               block_scan(thread_data, thread_data)
               for i in range(4):
                   output_arr[tid * 4 + i] = thread_data[i]

    :param dtype: Data type of the input and output values.
    :type dtype: DtypeType

    :param threads_per_block: Number of threads in the block. Can be an
        integer for 1D blocks or a tuple of two or three integers for 2D
        and 3D blocks.
    :type threads_per_block: DimType

    :param items_per_thread: Number of items owned by each thread.
        Must be greater than or equal to 1.
    :type items_per_thread: int, optional

    :param initial_value: Optional initial value for scan variants that
        support it.
    :type initial_value: Any, optional

    :param mode: Scan mode. Must be ``"exclusive"`` or ``"inclusive"``.
    :type mode: Literal["exclusive", "inclusive"], optional

    :param scan_op: Scan operator. The default is ``"+"``.
    :type scan_op: ScanOpType, optional

    :param block_prefix_callback_op: Optional block prefix callback
        operator invoked by the first warp.
    :type block_prefix_callback_op: Callable, optional

    :param algorithm: Scan algorithm. Must be ``"raking"``,
        ``"raking_memoize"``, or ``"warp_scans"``.
    :type algorithm:
        Literal["raking", "raking_memoize", "warp_scans"], optional

    :param methods: Optional method dictionary used for user-defined
        types.
    :type methods: dict, optional

    :raises ValueError: If ``algorithm`` is unsupported.
    :raises ValueError: If ``items_per_thread < 1``.
    :raises ValueError: If ``mode`` is not ``"exclusive"`` or
        ``"inclusive"``.
    :raises ValueError: If ``scan_op`` is unsupported.
    :raises ValueError: If ``initial_value`` is provided for sum scans.
    :raises ValueError: If ``initial_value`` is used with inclusive scans
        and ``items_per_thread == 1``.
    :raises ValueError: If ``initial_value`` is used with exclusive scans
        and ``items_per_thread == 1`` while
        ``block_prefix_callback_op`` is provided.
    :raises ValueError: If an initial value is required but cannot be
        inferred from ``dtype``.

    :returns: Callable primitive object that can be linked to and
        invoked from a CUDA kernel.
    :rtype: Callable
    """
    if algorithm not in CUB_BLOCK_SCAN_ALGOS:
        raise ValueError(f"Unsupported algorithm: {algorithm}")

    if items_per_thread < 1:
        raise ValueError("items_per_thread must be greater than or equal to 1")

    dim = normalize_dim_param(threads_per_block)
    dtype = normalize_dtype_param(dtype)

    if mode == "exclusive":
        cpp_func_prefix = "Exclusive"
    else:
        if mode != "inclusive":
            raise ValueError(f"Unsupported mode: {mode}")
        cpp_func_prefix = "Inclusive"

    # This will raise an error if scan_op is invalid.
    scan_op = ScanOp(scan_op)
    if scan_op.is_sum:
        # Make sure we specialize the correct CUB API for exclusive sum.
        cpp_function_name = f"{cpp_func_prefix}Sum"
    else:
        cpp_function_name = f"{cpp_func_prefix}Scan"

    # An initial value is not supported for inclusive and exclusive sums.
    if initial_value is not None and scan_op.is_sum:
        raise ValueError(
            "initial_value is not supported for inclusive and exclusive sums"
        )

    # An initial value is not supported for inclusive scans with a single
    # item per thread.
    invalid_initial_value = (
        initial_value is not None and items_per_thread == 1 and mode == "inclusive"
    )
    if invalid_initial_value:
        raise ValueError(
            "initial_value is not supported for inclusive scans with "
            "items_per_thread == 1"
        )

    # An initial value is not supported for exclusive scans with a
    # single item per thread and a block prefix callback operator.
    invalid_initial_value = (
        items_per_thread == 1
        and initial_value is not None
        and mode == "exclusive"
        and block_prefix_callback_op is not None
    )
    if invalid_initial_value:
        raise ValueError(
            "initial_value is not supported for exclusive scans with "
            "items_per_thread == 1 and a block prefix callback operator"
        )

    # An initial value is required for both inclusive and exclusive scans
    # when items_per_thread > 1 and a block prefix callback operator is
    # not supplied by the caller.
    initial_value_required = items_per_thread > 1 and block_prefix_callback_op is None
    if initial_value_required and initial_value is None:
        # We require an initial value, but one was not supplied.
        # Attempt to create a default value for the given dtype.
        # If we can't, raise an error.
        try:
            initial_value = dtype.cast_python_value(0)
        except (TypeError, NotImplementedError) as e:
            # We can't create a default value for the given dtype.
            # Raise an error.
            msg = (
                "initial_value is required for both inclusive and "
                "exclusive scans when items_per_thread > 1 and no "
                "block prefix callback operator has been supplied; "
                "attempted to create a default value for the given "
                f"dtype, but failed: {e}"
            )
            raise ValueError(msg) from e

    specialization_kwds = {
        "T": dtype,
        "BLOCK_DIM_X": dim[0],
        "ALGORITHM": CUB_BLOCK_SCAN_ALGOS[algorithm],
        "BLOCK_DIM_Y": dim[1],
        "BLOCK_DIM_Z": dim[2],
    }

    template_parameters = [
        TemplateParameter("T"),
        TemplateParameter("BLOCK_DIM_X"),
        TemplateParameter("ALGORITHM"),
        TemplateParameter("BLOCK_DIM_Y"),
        TemplateParameter("BLOCK_DIM_Z"),
    ]

    if items_per_thread == 1:
        fake_return = True
    else:
        specialization_kwds["ITEMS_PER_THREAD"] = items_per_thread
        fake_return = False

    # A "known" scan op is the standard set of associative operators,
    # e.g. ::cuda::std::plus<>, etc.  A "callable" scan op is a Python
    # callable that has been furnished by the caller.  Thus, we need to
    # generate different parameters for each case.
    if scan_op.is_known:

        def make_dependent_scan_op():
            return DependentCxxOperator(
                dep=Dependency("T"),
                cpp=scan_op.op_cpp,
            )

    elif scan_op.is_callable:

        def make_dependent_scan_op():
            return DependentPythonOperator(
                ret_dtype=Dependency("T"),
                arg_dtypes=[Dependency("T"), Dependency("T")],
                op=Dependency("ScanOp"),
            )

    if block_prefix_callback_op is not None:

        def make_dependent_block_prefix_callback_op():
            return DependentPythonOperator(
                ret_dtype=Dependency("T"),
                arg_dtypes=[Dependency("T")],
                op=Dependency("BlockPrefixCallbackOp"),
            )

    if scan_op.is_sum:
        if items_per_thread == 1:
            if block_prefix_callback_op is None:
                parameters = [
                    # Signature:
                    # void BlockScan<T, BLOCK_DIM_X, ALGORITHM,
                    #                   BLOCK_DIM_Y, BLOCK_DIM_Z>(
                    #     temp_storage
                    # ).<Inclusive|Exclusive>Sum(
                    #     T, # input
                    #     T& # output
                    # )
                    [
                        # temp_storage
                        Pointer(numba.uint8),
                        # T input
                        DependentReference(Dependency("T")),
                        # T& output
                        DependentReference(Dependency("T"), is_output=True),
                    ],
                ]
            else:
                parameters = [
                    # Signature:
                    # void BlockScan<T, BLOCK_DIM_X, ALGORITHM,
                    #                   BLOCK_DIM_Y, BLOCK_DIM_Z>(
                    #     temp_storage
                    # ).<Inclusive|Exclusive>Sum(
                    #     T,                     # input
                    #     T&,                    # output
                    #     BlockPrefixCallbackOp& # block_prefix_callback_op
                    # )
                    [
                        # temp_storage
                        Pointer(numba.uint8),
                        # T input
                        DependentReference(Dependency("T")),
                        # T& output
                        DependentReference(Dependency("T"), is_output=True),
                        # BlockPrefixCallbackOp& block_prefix_callback_op
                        make_dependent_block_prefix_callback_op(),
                    ],
                ]

        else:
            assert items_per_thread > 1, items_per_thread
            if block_prefix_callback_op is not None:
                parameters = [
                    # Signature:
                    # void BlockScan<T, BLOCK_DIM_X, ALGORITHM,
                    #                   BLOCK_DIM_Y, BLOCK_DIM_Z>(
                    #     temp_storage
                    # ).<Inclusive|Exclusive>Sum(
                    #     T (&)[ITEMS_PER_THREAD], # input
                    #     T (&)[ITEMS_PER_THREAD], # output
                    #     BlockPrefixCallbackOp&   # block_prefix_callback_op
                    # )
                    [
                        # temp_storage
                        Pointer(numba.uint8),
                        # T (&)[ITEMS_PER_THREAD] input
                        DependentArray(Dependency("T"), Dependency("ITEMS_PER_THREAD")),
                        # T (&)[ITEMS_PER_THREAD] output
                        DependentArray(Dependency("T"), Dependency("ITEMS_PER_THREAD")),
                        # BlockPrefixCallbackOp& block_prefix_callback_op
                        make_dependent_block_prefix_callback_op(),
                    ],
                ]
            else:
                parameters = [
                    # Signature:
                    # void BlockScan<T, BLOCK_DIM_X, ALGORITHM,
                    #                   BLOCK_DIM_Y, BLOCK_DIM_Z>(
                    #     temp_storage
                    # ).<Inclusive|Exclusive>Sum(
                    #     T (&)[ITEMS_PER_THREAD], # input
                    #     T (&)[ITEMS_PER_THREAD]  # output
                    # )
                    [
                        # temp_storage
                        Pointer(numba.uint8),
                        # T (&)[ITEMS_PER_THREAD] input
                        DependentArray(Dependency("T"), Dependency("ITEMS_PER_THREAD")),
                        # T (&)[ITEMS_PER_THREAD] output
                        DependentArray(Dependency("T"), Dependency("ITEMS_PER_THREAD")),
                    ],
                ]

    elif scan_op.is_known or scan_op.is_callable:
        if items_per_thread == 1:
            if mode == "exclusive":
                if block_prefix_callback_op is not None:
                    assert initial_value is None
                    parameters = [
                        # Signature:
                        # void BlockScan<T, BLOCK_DIM_X, ALGORITHM,
                        #                   BLOCK_DIM_Y, BLOCK_DIM_Z>(
                        #     temp_storage
                        # )<ScanOp>ExclusiveScan(
                        #     T,                     # input
                        #     T&,                    # output
                        #     ScanOp,                # scan_op
                        #     BlockPrefixCallbackOp& # block_prefix_callback_op
                        # )
                        [
                            # temp_storage
                            Pointer(numba.uint8),
                            # T input
                            DependentReference(Dependency("T")),
                            # T& output
                            DependentReference(Dependency("T"), is_output=True),
                            # ScanOp scan_op
                            make_dependent_scan_op(),
                            # BlockPrefixCallbackOp& block_prefix_callback_op
                            make_dependent_block_prefix_callback_op(),
                        ],
                    ]
                else:
                    parameters = [
                        # Signature:
                        # void BlockScan<T, BLOCK_DIM_X, ALGORITHM,
                        #                   BLOCK_DIM_Y, BLOCK_DIM_Z>(
                        #     temp_storage
                        # )<ScanOp>ExclusiveScan(
                        #     T,     # input
                        #     T&,    # output
                        #     ScanOp # scan_op
                        # )
                        [
                            # temp_storage
                            Pointer(numba.uint8),
                            # T input
                            DependentReference(Dependency("T")),
                            # T& output
                            DependentReference(Dependency("T"), is_output=True),
                            # ScanOp scan_op
                            make_dependent_scan_op(),
                        ],
                        # Signature:
                        # void BlockScan<T, BLOCK_DIM_X, ALGORITHM,
                        #                   BLOCK_DIM_Y, BLOCK_DIM_Z>(
                        #     temp_storage
                        # )<ScanOp>ExclusiveScan(
                        #     T,     # input
                        #     T&,    # output
                        #     T,     # initial_value
                        #     ScanOp # scan_op
                        # )
                        [
                            # temp_storage
                            Pointer(numba.uint8),
                            # T input
                            DependentReference(Dependency("T")),
                            # T& output
                            DependentReference(Dependency("T"), is_output=True),
                            # T initial_value
                            DependentReference(Dependency("T")),
                            # ScanOp scan_op
                            make_dependent_scan_op(),
                        ],
                    ]

            else:
                assert mode == "inclusive" or initial_value is None
                if block_prefix_callback_op is not None:
                    parameters = [
                        # Signature:
                        # void BlockScan<T, BLOCK_DIM_X, ALGORITHM,
                        #                   BLOCK_DIM_Y, BLOCK_DIM_Z>(
                        #     temp_storage
                        # )<ScanOp><Inclusive|Exclusive>Scan(
                        #     T,                     # input
                        #     T&,                    # output
                        #     ScanOp,                # scan_op
                        #     BlockPrefixCallbackOp& # block_prefix_callback_op
                        # )
                        [
                            # temp_storage
                            Pointer(numba.uint8),
                            # T input
                            DependentReference(Dependency("T")),
                            # T& output
                            DependentReference(Dependency("T"), is_output=True),
                            # ScanOp scan_op
                            make_dependent_scan_op(),
                            # BlockPrefixCallbackOp& block_prefix_callback_op
                            make_dependent_block_prefix_callback_op(),
                        ],
                    ]
                else:
                    parameters = [
                        # Signature:
                        # void BlockScan<T, BLOCK_DIM_X, ALGORITHM,
                        #                   BLOCK_DIM_Y, BLOCK_DIM_Z>(
                        #     temp_storage
                        # )<ScanOp><Inclusive|Exclusive>Scan(
                        #     T,     # input
                        #     T&,    # output
                        #     ScanOp # scan_op
                        # )
                        [
                            # temp_storage
                            Pointer(numba.uint8),
                            # T input
                            DependentReference(Dependency("T")),
                            # T& output
                            DependentReference(Dependency("T"), is_output=True),
                            # ScanOp scan_op
                            make_dependent_scan_op(),
                        ],
                    ]

        else:
            assert items_per_thread > 1, items_per_thread
            if block_prefix_callback_op is not None:
                assert initial_value is None
                parameters = [
                    # Signature:
                    # void BlockScan<T, BLOCK_DIM_X, ALGORITHM,
                    #                   BLOCK_DIM_Y, BLOCK_DIM_Z>(
                    #     temp_storage
                    # )<ITEMS_PER_THREAD, ScanOp><Inclusive|Exclusive>Scan(
                    #     T (&)[ITEMS_PER_THREAD], # input
                    #     T (&)[ITEMS_PER_THREAD], # output
                    #     ScanOp,                  # scan_op
                    #     BlockPrefixCallbackOp&   # block_prefix_callback_op
                    # )
                    [
                        # temp_storage
                        Pointer(numba.uint8),
                        # T (&)[ITEMS_PER_THREAD] input
                        DependentArray(Dependency("T"), Dependency("ITEMS_PER_THREAD")),
                        # T (&)[ITEMS_PER_THREAD] output
                        DependentArray(Dependency("T"), Dependency("ITEMS_PER_THREAD")),
                        # ScanOp scan_op
                        make_dependent_scan_op(),
                        # BlockPrefixCallbackOp& block_prefix_callback_op
                        make_dependent_block_prefix_callback_op(),
                    ],
                ]
            else:
                parameters = [
                    # Signature:
                    # void BlockScan<T, BLOCK_DIM_X, ALGORITHM,
                    #                   BLOCK_DIM_Y, BLOCK_DIM_Z>(
                    #     temp_storage
                    # )<ITEMS_PER_THREAD, ScanOp><Inclusive|Exclusive>Scan(
                    #     T (&)[ITEMS_PER_THREAD], # input
                    #     T (&)[ITEMS_PER_THREAD], # output
                    #     ScanOp                   # scan_op
                    # )
                    [
                        # temp_storage
                        Pointer(numba.uint8),
                        # T (&)[ITEMS_PER_THREAD] input
                        DependentArray(Dependency("T"), Dependency("ITEMS_PER_THREAD")),
                        # T (&)[ITEMS_PER_THREAD] output
                        DependentArray(Dependency("T"), Dependency("ITEMS_PER_THREAD")),
                        # ScanOp scan_op
                        make_dependent_scan_op(),
                    ],
                    # Signature:
                    # void BlockScan<T, BLOCK_DIM_X, ALGORITHM,
                    #                   BLOCK_DIM_Y, BLOCK_DIM_Z>(
                    #     temp_storage
                    # )<ITEMS_PER_THREAD, ScanOp><Inclusive|Exclusive>Scan(
                    #     T (&)[ITEMS_PER_THREAD], # input
                    #     T (&)[ITEMS_PER_THREAD], # output
                    #     T,                       # initial_value
                    #     ScanOp                   # scan_op
                    # )
                    [
                        # temp_storage
                        Pointer(numba.uint8),
                        # T (&)[ITEMS_PER_THREAD] input
                        DependentArray(Dependency("T"), Dependency("ITEMS_PER_THREAD")),
                        # T (&)[ITEMS_PER_THREAD] output
                        DependentArray(Dependency("T"), Dependency("ITEMS_PER_THREAD")),
                        # T initial_value
                        DependentReference(Dependency("T")),
                        # ScanOp scan_op
                        make_dependent_scan_op(),
                    ],
                ]

    else:
        # We shouldn't ever hit this; there are only three types of scan ops
        # supported: sum, known, and callable.
        raise RuntimeError("Unreachable code")

    # Invariant check: if we get here, parameters shouldn't be empty.
    assert parameters, "parameters should not be empty"

    # If we have a non-None `methods`, we're dealing with user-defined types.
    if methods is not None:
        type_definitions = [
            numba_type_to_wrapper(dtype, methods=methods),
        ]
    else:
        type_definitions = None

    template = Algorithm(
        "BlockScan",
        cpp_function_name,
        "block_scan",
        ["cub/block/block_scan.cuh"],
        template_parameters,
        parameters,
        fake_return=fake_return,
        type_definitions=type_definitions,
    )

    if scan_op.is_callable:
        specialization_kwds["ScanOp"] = scan_op.op

    if block_prefix_callback_op is not None:
        specialization_kwds["BlockPrefixCallbackOp"] = block_prefix_callback_op

    specialization = template.specialize(specialization_kwds)
    return Invocable(
        temp_files=[
            make_binary_tempfile(ltoir, ".ltoir")
            for ltoir in specialization.get_lto_ir()
        ],
        temp_storage_bytes=specialization.temp_storage_bytes,
        temp_storage_alignment=specialization.temp_storage_alignment,
        algorithm=specialization,
    )


def make_exclusive_sum(
    dtype: DtypeType,
    threads_per_block: DimType,
    items_per_thread: int,
    prefix_op: Callable = None,
    algorithm: Literal["raking", "raking_memoize", "warp_scans"] = "raking",
    methods: dict = None,
) -> Callable:
    """
    Creates an exclusive block-wide prefix sum primitive using addition
    (+) as the scan operator.

    Example:
        The code snippet below illustrates an exclusive prefix sum of
        512 integer items in a
        :ref:`blocked arrangement <flexible-data-arrangement>` across
        128 threads where each thread owns 4 consecutive items.

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_block_scan_api.py
            :language: python
            :dedent:
            :start-after: example-begin imports
            :end-before: example-end imports

        The following snippet shows how to invoke the returned
        ``block_exclusive_sum`` primitive:

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_block_scan_api.py
            :language: python
            :dedent:
            :start-after: example-begin exclusive-sum
            :end-before: example-end exclusive-sum

        Suppose the set of input ``thread_data`` across the block of
        threads is
        ``{ [1, 1, 1, 1], [1, 1, 1, 1], ..., [1, 1, 1, 1] }``.

        The corresponding output ``thread_data`` in those threads will be
        ``{ [0, 1, 2, 3], [4, 5, 6, 7], ..., [508, 509, 510, 511] }``.

    :param dtype: Data type of the input and output values.
    :type dtype: DtypeType
    :param threads_per_block: Number of threads in the block.
    :type threads_per_block: DimType
    :param items_per_thread: Number of items owned by each thread.
    :type items_per_thread: int, optional
    :param prefix_op: Optional block prefix callback operator.
    :type prefix_op: Callable, optional
    :param algorithm: Scan algorithm.
    :type algorithm:
        Literal["raking", "raking_memoize", "warp_scans"], optional
    :param methods: Optional method dictionary for user-defined types.
    :type methods: dict, optional
    :returns: Callable primitive object for exclusive prefix sum.
    :rtype: Callable
    """
    return make_scan(
        dtype=dtype,
        threads_per_block=threads_per_block,
        items_per_thread=items_per_thread,
        mode="exclusive",
        scan_op="+",
        block_prefix_callback_op=prefix_op,
        algorithm=algorithm,
        methods=methods,
    )


def make_inclusive_sum(
    dtype: DtypeType,
    threads_per_block: DimType,
    items_per_thread: int,
    prefix_op: Callable = None,
    algorithm: Literal["raking", "raking_memoize", "warp_scans"] = "raking",
    methods: dict = None,
) -> Callable:
    """
    Creates an inclusive block-wide prefix sum primitive using addition
    (+) as the scan operator.

    Example:
        The snippet below shows how to create and invoke the returned
        ``block_inclusive_sum`` primitive.

        .. code-block:: python

           block_inclusive_sum = coop.block.make_inclusive_sum(
               dtype=numba.int32,
               threads_per_block=128,
               items_per_thread=4,
           )

           @cuda.jit(link=block_inclusive_sum.files)
           def kernel(thread_data):
               block_inclusive_sum(thread_data, thread_data)

    :param dtype: Data type of the input and output values.
    :type dtype: DtypeType
    :param threads_per_block: Number of threads in the block.
    :type threads_per_block: DimType
    :param items_per_thread: Number of items owned by each thread.
    :type items_per_thread: int, optional
    :param prefix_op: Optional block prefix callback operator.
    :type prefix_op: Callable, optional
    :param algorithm: Scan algorithm.
    :type algorithm:
        Literal["raking", "raking_memoize", "warp_scans"], optional
    :param methods: Optional method dictionary for user-defined types.
    :type methods: dict, optional
    :returns: Callable primitive object for inclusive prefix sum.
    :rtype: Callable
    """
    return make_scan(
        dtype=dtype,
        threads_per_block=threads_per_block,
        items_per_thread=items_per_thread,
        mode="inclusive",
        scan_op="+",
        block_prefix_callback_op=prefix_op,
        algorithm=algorithm,
        methods=methods,
    )


def make_exclusive_scan(
    dtype: DtypeType,
    threads_per_block: DimType,
    scan_op: ScanOpType,
    items_per_thread: int,
    initial_value: Any = None,
    prefix_op: Callable = None,
    algorithm: Literal["raking", "raking_memoize", "warp_scans"] = "raking",
    methods: dict = None,
) -> Callable:
    """
    Creates an exclusive block-wide prefix scan primitive with the
    specified scan operator.

    Example:
        The snippet below shows how to create and invoke the returned
        ``block_exclusive_scan`` primitive.

        .. code-block:: python

           block_exclusive_scan = coop.block.make_exclusive_scan(
               dtype=numba.int32,
               threads_per_block=128,
               scan_op="max",
               items_per_thread=4,
           )

           @cuda.jit(link=block_exclusive_scan.files)
           def kernel(thread_data):
               block_exclusive_scan(thread_data, thread_data)

    :param dtype: Data type of the input and output values.
    :type dtype: DtypeType
    :param threads_per_block: Number of threads in the block.
    :type threads_per_block: DimType
    :param scan_op: Scan operator.
    :type scan_op: ScanOpType
    :param items_per_thread: Number of items owned by each thread.
    :type items_per_thread: int, optional
    :param initial_value: Optional initial value when supported.
    :type initial_value: Any, optional
    :param prefix_op: Optional block prefix callback operator.
    :type prefix_op: Callable, optional
    :param algorithm: Scan algorithm.
    :type algorithm:
        Literal["raking", "raking_memoize", "warp_scans"], optional
    :param methods: Optional method dictionary for user-defined types.
    :type methods: dict, optional
    :returns: Callable primitive object for exclusive prefix scan.
    :rtype: Callable
    """
    return make_scan(
        dtype=dtype,
        threads_per_block=threads_per_block,
        items_per_thread=items_per_thread,
        initial_value=initial_value,
        mode="exclusive",
        scan_op=scan_op,
        block_prefix_callback_op=prefix_op,
        algorithm=algorithm,
        methods=methods,
    )


def make_inclusive_scan(
    dtype: DtypeType,
    threads_per_block: DimType,
    scan_op: ScanOpType,
    items_per_thread: int,
    initial_value: Any = None,
    prefix_op: Callable = None,
    algorithm: Literal["raking", "raking_memoize", "warp_scans"] = "raking",
    methods: dict = None,
) -> Callable:
    """
    Creates an inclusive block-wide prefix scan primitive with the
    specified scan operator.

    Example:
        The snippet below shows how to create and invoke the returned
        ``block_inclusive_scan`` primitive.

        .. code-block:: python

           block_inclusive_scan = coop.block.make_inclusive_scan(
               dtype=numba.int32,
               threads_per_block=128,
               scan_op="min",
               items_per_thread=4,
           )

           @cuda.jit(link=block_inclusive_scan.files)
           def kernel(thread_data):
               block_inclusive_scan(thread_data, thread_data)

    :param dtype: Data type of the input and output values.
    :type dtype: DtypeType
    :param threads_per_block: Number of threads in the block.
    :type threads_per_block: DimType
    :param scan_op: Scan operator.
    :type scan_op: ScanOpType
    :param items_per_thread: Number of items owned by each thread.
    :type items_per_thread: int, optional
    :param initial_value: Optional initial value when supported.
    :type initial_value: Any, optional
    :param prefix_op: Optional block prefix callback operator.
    :type prefix_op: Callable, optional
    :param algorithm: Scan algorithm.
    :type algorithm:
        Literal["raking", "raking_memoize", "warp_scans"], optional
    :param methods: Optional method dictionary for user-defined types.
    :type methods: dict, optional
    :returns: Callable primitive object for inclusive prefix scan.
    :rtype: Callable
    """
    return make_scan(
        dtype=dtype,
        threads_per_block=threads_per_block,
        items_per_thread=items_per_thread,
        initial_value=initial_value,
        mode="inclusive",
        scan_op=scan_op,
        block_prefix_callback_op=prefix_op,
        algorithm=algorithm,
        methods=methods,
    )



# Copyright (c) 2024-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional, Tuple

import numpy as np

"""
Utilities for extracting information from protocols such as `__cuda_array_interface__` and `__cuda_stream__`.
"""

if TYPE_CHECKING:
    from ..typing import DeviceArrayLike, GpuStruct


def is_device_array(obj: object) -> bool:
    """Check if an object implements the `__cuda_array_interface__` protocol."""
    return hasattr(obj, "__cuda_array_interface__")


def get_data_pointer(arr: DeviceArrayLike) -> int:
    # TODO: these are fast paths for CuPy and PyTorch until
    # we have a more general solution.

    # Fast path for PyTorch (arr.data_ptr())
    try:
        return arr.data_ptr()  # type: ignore
    except AttributeError:
        pass

    # Fast path for CuPy (arr.data.ptr)
    try:
        return arr.data.ptr  # type: ignore
    except AttributeError:
        pass

    # Fall back to __cuda_array_interface__
    return arr.__cuda_array_interface__["data"][0]


def get_dtype(arr: DeviceArrayLike | GpuStruct | np.ndarray) -> np.dtype:
    # Try the fast path via .dtype attribute (works for np.ndarray, GpuStruct, and most device arrays)
    try:
        return np.dtype(arr.dtype)  # type: ignore
    except (AttributeError, TypeError):
        pass

    # Fall back to __cuda_array_interface__ for DeviceArrayLike
    cai = arr.__cuda_array_interface__  # type: ignore
    typestr = cai["typestr"]

    if typestr.startswith("|V"):
        # it's a structured dtype, use the descr field:
        return np.dtype(cai["descr"])
    else:
        # a simple dtype, use the typestr field:
        return np.dtype(typestr)


def get_shape(arr: DeviceArrayLike) -> Tuple[int]:
    try:
        # TODO: this is a fast path for CuPy until
        # we have a more general solution.
        return arr.shape  # type: ignore
    except AttributeError:
        return arr.__cuda_array_interface__["shape"]


def get_size(arr: DeviceArrayLike) -> int:
    """Get the total number of elements in an array."""
    # Try fast path via .size attribute
    try:
        return int(arr.size)  # type: ignore
    except AttributeError:
        pass

    # Fall back to computing from shape
    shape = get_shape(arr)
    import math

    return math.prod(shape)


def is_contiguous(arr: DeviceArrayLike) -> bool:
    cai = arr.__cuda_array_interface__

    strides = cai["strides"]

    if strides is None:
        return True

    shape = cai["shape"]

    if any(dim == 0 for dim in shape):
        # array has no elements
        return True

    if all(dim == 1 for dim in shape):
        # there is a single element:
        return True

    itemsize = get_dtype(arr).itemsize

    if strides[-1] == itemsize:
        # assume C-contiguity
        expected_stride = itemsize
        for dim, stride in zip(reversed(shape), reversed(strides)):
            if stride != expected_stride:
                return False
            expected_stride *= dim
        return True
    elif strides[0] == itemsize:
        # assume F-contiguity
        expected_stride = itemsize
        for dim, stride in zip(shape, strides):
            if stride != expected_stride:
                return False
            expected_stride *= dim
        return True
    else:
        # not contiguous
        return False


def compute_c_contiguous_strides_in_bytes(
    shape: Tuple[int], itemsize: int
) -> Tuple[int, ...]:
    """Return C-contiguous strides in bytes for a given shape and itemsize (compatible with NumPy .strides)."""

    strides: List[int] = []
    acc = itemsize

    for dim in reversed(shape):
        strides.insert(0, acc)
        acc *= dim

    return tuple(strides)


def validate_and_get_stream(stream) -> Optional[int]:
    # null stream is allowed
    if stream is None:
        return None

    try:
        stream_property = stream.__cuda_stream__()
    except AttributeError as e:
        raise TypeError(
            f"stream argument {stream} does not implement the '__cuda_stream__' protocol"
        ) from e

    try:
        version, handle, *_ = stream_property
    except (TypeError, ValueError) as e:
        raise TypeError(
            f"could not obtain __cuda_stream__ protocol version and handle from {stream_property}"
        ) from e

    if version == 0:
        if not isinstance(handle, int):
            raise TypeError(f"invalid stream handle {handle}")
        return handle

    raise TypeError(f"unsupported __cuda_stream__ version {version}")



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from __future__ import annotations

import numpy as np

from .. import _bindings, types
from .. import _cccl_interop as cccl
from .._caching import cache_with_registered_key_functions
from .._cccl_interop import call_build, set_cccl_iterator_state
from .._utils import protocols
from ..op import OpAdapter, OpKind, make_op_adapter
from ..typing import DeviceArrayLike, IteratorT, Operator


def _normalize_comp(comp: Operator | None) -> OpAdapter:
    # Use a lambda for the default comparator rather than OpKind.LESS
    # because well-known ops don't carry type information needed by
    # the binary search JIT compilation.
    if comp is None or comp is OpKind.LESS:

        def _default_less(a, b):
            return a < b

        return make_op_adapter(_default_less)
    return make_op_adapter(comp)


class _BinarySearch:
    __slots__ = [
        "build_result",
        "d_data_cccl",
        "d_values_cccl",
        "d_out_cccl",
        "op_cccl",
        "data_ptr",
        "out_ptr",
    ]

    def __init__(
        self,
        d_data: DeviceArrayLike,
        d_values: DeviceArrayLike | IteratorT,
        d_out: DeviceArrayLike,
        comp: OpAdapter,
        mode: _bindings.BinarySearchMode,
    ):
        if not protocols.is_device_array(d_data):
            raise ValueError("d_data must be a device array for index outputs.")
        if not protocols.is_device_array(d_out):
            raise ValueError("d_out must be a device array for index outputs.")

        out_dtype = protocols.get_dtype(d_out)
        if out_dtype.kind != "u":
            raise TypeError("d_out must use an unsigned integer dtype for indices.")
        if out_dtype.itemsize != np.dtype(np.uintp).itemsize:
            raise ValueError(
                "d_out must use a pointer-sized unsigned integer dtype (np.uintp)."
            )

        self.data_ptr = protocols.get_data_pointer(d_data)
        self.out_ptr = protocols.get_data_pointer(d_out)

        self.d_data_cccl = cccl.to_cccl_input_iter(d_data)
        self.d_values_cccl = cccl.to_cccl_input_iter(d_values)
        data_value_type = cccl.get_value_type(d_data)
        self.d_out_cccl = cccl.to_cccl_output_iter(d_out)

        self.op_cccl = comp.compile((data_value_type, data_value_type), types.uint8)

        self.build_result = call_build(
            _bindings.DeviceBinarySearchBuildResult,
            mode,
            self.d_data_cccl,
            self.d_values_cccl,
            self.d_out_cccl,
            self.op_cccl,
        )

    def __call__(
        self,
        d_data,
        d_values,
        d_out,
        comp: Operator | None,
        num_items: int,
        num_values: int,
        stream=None,
    ):
        set_cccl_iterator_state(self.d_data_cccl, d_data)
        set_cccl_iterator_state(self.d_values_cccl, d_values)
        set_cccl_iterator_state(self.d_out_cccl, d_out)

        # Update op state for stateful ops
        comp_adapter = (
            _normalize_comp(comp) if comp is not None else _normalize_comp(None)
        )
        self.op_cccl.state = comp_adapter.get_state()

        stream_handle = protocols.validate_and_get_stream(stream)
        self.build_result.compute(
            self.d_data_cccl,
            num_items,
            self.d_values_cccl,
            num_values,
            self.d_out_cccl,
            self.op_cccl,
            stream_handle,
        )


@cache_with_registered_key_functions
def _make_binary_search(
    d_data: DeviceArrayLike,
    d_values: DeviceArrayLike | IteratorT,
    d_out: DeviceArrayLike,
    comp: OpAdapter,
    mode: _bindings.BinarySearchMode,
    data_ptr: int,
    out_ptr: int,
):
    """Cached factory for _BinarySearch."""
    return _BinarySearch(d_data, d_values, d_out, comp, mode)


def make_lower_bound(
    d_data: DeviceArrayLike,
    d_values: DeviceArrayLike | IteratorT,
    d_out: DeviceArrayLike,
    comp: Operator | None = None,
):
    """
    Create a lower_bound object that can be called to find insertion positions.

    Example:
        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/binary_search/lower_bound_object.py
            :language: python
            :start-after: # example-begin

    Args:
        d_data: Device array containing the sorted input range.
        d_values: Device array or iterator containing the search values.
        d_out: Device array to store the index results.
        comp: Optional comparison operator (default: ``OpKind.LESS``).

    Returns:
        A callable object that performs lower_bound.

    See Also:
        :func:`lower_bound`
    """
    comp_adapter = _normalize_comp(comp)
    return _make_binary_search(
        d_data,
        d_values,
        d_out,
        comp_adapter,
        _bindings.BinarySearchMode.LOWER_BOUND,
        protocols.get_data_pointer(d_data),
        protocols.get_data_pointer(d_out),
    )


def make_upper_bound(
    d_data: DeviceArrayLike,
    d_values: DeviceArrayLike | IteratorT,
    d_out: DeviceArrayLike,
    comp: Operator | None = None,
):
    """
    Create an upper_bound object that can be called to find insertion positions.

    Example:
        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/binary_search/upper_bound_object.py
            :language: python
            :start-after: # example-begin

    Args:
        d_data: Device array containing the sorted input range.
        d_values: Device array or iterator containing the search values.
        d_out: Device array to store the index results.
        comp: Optional comparison operator (default: ``OpKind.LESS``).

    Returns:
        A callable object that performs upper_bound.

    See Also:
        :func:`upper_bound`
    """
    comp_adapter = _normalize_comp(comp)
    return _make_binary_search(
        d_data,
        d_values,
        d_out,
        comp_adapter,
        _bindings.BinarySearchMode.UPPER_BOUND,
        protocols.get_data_pointer(d_data),
        protocols.get_data_pointer(d_out),
    )


def lower_bound(
    d_data: DeviceArrayLike,
    d_values: DeviceArrayLike | IteratorT,
    d_out: DeviceArrayLike,
    num_items: int,
    num_values: int,
    comp: Operator | None = None,
    stream=None,
):
    """
    Find the *first* position that each value in ``d_values`` would be inserted into
    ``d_data`` to maintain sorted order.

    Example:
        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/binary_search/lower_bound_basic.py
            :language: python
            :start-after: # example-begin

    Args:
        d_data: Device array containing the sorted input range.
        d_values: Device array or iterator containing the search values.
        d_out: Device array to store the index results.
        num_items: Number of items in ``d_data``.
        num_values: Number of items in ``d_values``.
        comp: Optional comparison operator (default: ``OpKind.LESS``).
        stream: CUDA stream for the operation (optional).
    """
    searcher = make_lower_bound(d_data, d_values, d_out, comp)
    searcher(d_data, d_values, d_out, comp, num_items, num_values, stream)


def upper_bound(
    d_data: DeviceArrayLike,
    d_values: DeviceArrayLike | IteratorT,
    d_out: DeviceArrayLike,
    num_items: int,
    num_values: int,
    comp: Operator | None = None,
    stream=None,
):
    """
    Find the *last* position that each value in ``d_values`` would be inserted into
    ``d_data`` to maintain sorted order.

    Example:
        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/binary_search/upper_bound_basic.py
            :language: python
            :start-after: # example-begin

    Args:
        d_data: Device array containing the sorted input range.
        d_values: Device array or iterator containing the search values.
        d_out: Device array to store the index results.
        num_items: Number of items in ``d_data``.
        num_values: Number of items in ``d_values``.
        comp: Optional comparison operator (default: ``OpKind.LESS``).
        stream: CUDA stream for the operation (optional).
    """
    searcher = make_upper_bound(d_data, d_values, d_out, comp)
    searcher(d_data, d_values, d_out, comp, num_items, num_values, stream)



# Copyright (c) 2025, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import sys
from dataclasses import dataclass
from functools import lru_cache
from importlib.resources import as_file, files
from pathlib import Path
from typing import Optional

# type: ignore[import-not-found]
from cuda.pathfinder import find_nvidia_header_directory


@dataclass
class IncludePaths:
    cuda: Optional[Path]
    libcudacxx: Optional[Path]
    cub: Optional[Path]
    thrust: Optional[Path]

    def as_tuple(self):
        # Note: higher-level ... lower-level order:
        return (self.thrust, self.cub, self.libcudacxx, self.cuda)


@lru_cache()
def get_include_paths(probe_file: str = "cub/version.cuh") -> IncludePaths:
    cuda_incl = find_nvidia_header_directory("cudart")
    if cuda_incl is None:
        raise RuntimeError("Unable to locate CUDA include directory.")

    with as_file(files("cuda.cccl.headers.include")) as f:
        cccl_incl = Path(f)

    probe_file_path = Path(probe_file)
    if not (cccl_incl / probe_file_path).exists():
        for sp in sys.path:
            cccl_incl = Path(sp).resolve() / "cuda" / "cccl" / "headers" / "include"
            if (cccl_incl / probe_file_path).exists():
                break
        else:
            raise RuntimeError("Unable to locate CCCL include directory.")

    return IncludePaths(
        cuda=cuda_incl,
        libcudacxx=cccl_incl,
        cub=cccl_incl,
        thrust=cccl_incl,
    )



# Copyright (c) 2024-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
from __future__ import annotations

import enum
import functools
import os
import subprocess
import tempfile
import warnings
from typing import Callable, List

try:
    from cuda.core import Device as CudaDevice
except ImportError:
    from cuda.core.experimental import Device as CudaDevice


import numpy as np

# TODO: adding a type-ignore here because `cuda` being a
# namespace package confuses mypy when `cuda.<something_else>`
# is installed, but not `cuda.cccl`. For namespace packages,
# it appears we need to actually install the sub-package
# in order for mypy to find its py.typed file. However, CI
# does type checking of `cuda.cccl` without actually installing
# it.
#
# We need to find a better solution for this.
from cuda.cccl import get_include_paths  # type: ignore

from . import types
from ._bindings import (
    CommonData,
    Iterator,
    IteratorKind,
    IteratorState,
    Op,
    OpKind,
    Pointer,
    TypeEnum,
    TypeInfo,
    Value,
    make_pointer_object,
)
from ._utils.protocols import get_data_pointer, get_dtype, is_contiguous
from .iterators._base import IteratorBase
from .typing import DeviceArrayLike, GpuStruct

# Mapping from numpy dtype to TypeEnum for creating TypeInfo
_NUMPY_DTYPE_TO_ENUM = {
    np.dtype("int8"): TypeEnum.INT8,
    np.dtype("int16"): TypeEnum.INT16,
    np.dtype("int32"): TypeEnum.INT32,
    np.dtype("int64"): TypeEnum.INT64,
    np.dtype("uint8"): TypeEnum.UINT8,
    np.dtype("uint16"): TypeEnum.UINT16,
    np.dtype("uint32"): TypeEnum.UINT32,
    np.dtype("uint64"): TypeEnum.UINT64,
    np.dtype("float16"): TypeEnum.FLOAT16,
    np.dtype("float32"): TypeEnum.FLOAT32,
    np.dtype("float64"): TypeEnum.FLOAT64,
    np.dtype("bool"): TypeEnum.BOOLEAN,
}


@functools.lru_cache(maxsize=256)
def _type_info_from_dtype(dtype: np.dtype) -> TypeInfo:
    """
    Create a TypeInfo from a numpy dtype.
    Handles both primitive types and structured dtypes.
    """
    dtype = np.dtype(dtype)

    # Handle structured dtypes
    if dtype.type == np.void and dtype.fields is not None:
        return TypeInfo(dtype.itemsize, dtype.alignment, TypeEnum.STORAGE)

    if dtype.kind == "c":
        return TypeInfo(dtype.itemsize, dtype.alignment, TypeEnum.STORAGE)

    # Fallback for any other type
    type_enum = _NUMPY_DTYPE_TO_ENUM.get(dtype, TypeEnum.STORAGE)
    return TypeInfo(dtype.itemsize, dtype.alignment, type_enum)


def _is_well_known_op(op: OpKind) -> bool:
    return isinstance(op, OpKind) and op not in (OpKind.STATELESS, OpKind.STATEFUL)


def _device_array_to_cccl_iter(array: DeviceArrayLike) -> Iterator:
    if not is_contiguous(array):
        raise ValueError("Non-contiguous arrays are not supported.")
    dtype = get_dtype(array)

    info = _type_info_from_dtype(dtype)
    state_info = _type_info_from_dtype(np.intp)
    return Iterator(
        state_info.alignment,
        IteratorKind.POINTER,
        Op(),
        Op(),
        info,
        # Note: this is slightly slower, but supports all ndarray-like objects
        # as long as they support CAI
        # TODO: switch to use gpumemoryview once it's ready
        state=get_data_pointer(array),
    )


def _none_to_cccl_iter() -> Iterator:
    # Any type could be used here, we just need to pass NULL.
    info = _type_info_from_dtype(np.uint8)
    return Iterator(info.alignment, IteratorKind.POINTER, Op(), Op(), info, state=None)


class _IteratorIO(enum.Enum):
    INPUT = 0
    OUTPUT = 1


def _to_cccl_iter(
    it: DeviceArrayLike | IteratorBase | None, io_kind: _IteratorIO
) -> Iterator:
    if it is None:
        return _none_to_cccl_iter()
    if isinstance(it, IteratorBase):
        return it.to_cccl_iter(io_kind == _IteratorIO.OUTPUT)
    return _device_array_to_cccl_iter(it)


def to_cccl_input_iter(array_or_iterator) -> Iterator:
    return _to_cccl_iter(array_or_iterator, _IteratorIO.INPUT)


def to_cccl_output_iter(array_or_iterator) -> Iterator:
    return _to_cccl_iter(array_or_iterator, _IteratorIO.OUTPUT)


def to_cccl_value_state(array_or_struct: np.ndarray | GpuStruct) -> memoryview:
    if isinstance(array_or_struct, np.ndarray):
        assert array_or_struct.flags.contiguous
        data = array_or_struct.data.cast("B")
        return data
    else:
        # it's a GpuStruct, use the array underlying it
        return to_cccl_value_state(array_or_struct._data)


def to_cccl_value(array_or_struct: np.ndarray | GpuStruct) -> Value:
    if isinstance(array_or_struct, np.ndarray):
        info = _type_info_from_dtype(array_or_struct.dtype)
        return Value(info, array_or_struct.data.cast("B"))
    else:
        # it's a GpuStruct, use the array underlying it
        return to_cccl_value(array_or_struct._data)


def set_cccl_value_state(cccl_value: Value, array_or_struct: np.ndarray | GpuStruct):
    """
    Set the state of a CCCL Value object from a numpy array or GpuStruct.

    Args:
        cccl_value: The CCCL Value binding object
        array_or_struct: The numpy array or GpuStruct to get the state from
    """
    cccl_value.state = to_cccl_value_state(array_or_struct)


def get_value_type(
    d_in: DeviceArrayLike | IteratorBase | GpuStruct | np.ndarray,
):
    from .struct import _Struct

    if isinstance(d_in, IteratorBase):
        return d_in.value_type

    if isinstance(d_in, _Struct):
        return type(d_in)._type_descriptor  # type: ignore[union-attr]

    dtype = get_dtype(d_in)

    if dtype.type == np.void:
        return types.from_numpy_dtype(dtype)

    return types.from_numpy_dtype(dtype)


def set_cccl_iterator_state(cccl_it: Iterator, input_it):
    if cccl_it.is_kind_pointer():
        ptr = get_data_pointer(input_it)
        ptr_obj = make_pointer_object(ptr, input_it)
        cccl_it.state = ptr_obj
    else:
        state_ = input_it.state
        if isinstance(state_, (IteratorState, Pointer)):
            cccl_it.state = state_
        else:
            cccl_it.state = make_pointer_object(state_, input_it)


@functools.lru_cache()
def get_includes() -> List[str]:
    def as_option(p):
        if p is None:
            return ""
        return f"-I{p}"

    paths = get_include_paths().as_tuple()
    opts = [as_option(path) for path in paths]
    return opts


def _check_compile_result(cubin: bytes):
    # check compiled code for LDL/STL instructions
    temp_cubin_file = tempfile.NamedTemporaryFile(delete=False)
    try:
        temp_cubin_file.write(cubin)
        out = subprocess.run(
            ["nvdisasm", "-gi", temp_cubin_file.name], capture_output=True
        )
        if out.returncode != 0:
            raise RuntimeError("nvdisasm failed")
        sass = out.stdout.decode("utf-8")
    except FileNotFoundError:
        sass = "nvdiasm not found, skipping SASS validation"
        warnings.warn(sass)

    assert "LDL" not in sass, "LDL instruction found in SASS"
    assert "STL" not in sass, "STL instruction found in SASS"
    return temp_cubin_file.name


# this global variable controls whether the compile result is checked
# for LDL/STL instructions. Should be set to `True` for testing only.
_check_sass: bool = False


def call_build(build_impl_fn: Callable, *args, **kwargs):
    """Calls given build_impl_fn callable while providing compute capability and paths

    Returns result of the call.
    """
    global _check_sass

    cc_major, cc_minor = CudaDevice().compute_capability
    cub_path, thrust_path, libcudacxx_path, cuda_include_path = get_includes()
    common_data = CommonData(
        cc_major, cc_minor, cub_path, thrust_path, libcudacxx_path, cuda_include_path
    )
    result = build_impl_fn(
        *args,
        common_data,
        **kwargs,
    )

    if _check_sass:
        cubin = result._get_cubin()
        temp_cubin_file_name = _check_compile_result(cubin)
        os.unlink(temp_cubin_file_name)

    return result



# Copyright (c) 2024-2025, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import re
import tempfile
from collections import namedtuple
from enum import Enum
from typing import BinaryIO, Union

import numba
import numpy as np

from ._typing import DimType

version = namedtuple("version", ("major", "minor"))
code = namedtuple("code", ("kind", "version", "data"))
symbol = namedtuple("symbol", ("kind", "name"))
dim3 = namedtuple("dim3", ("x", "y", "z"))


CUB_BLOCK_SCAN_ALGOS = {
    "raking": "::cub::BlockScanAlgorithm::BLOCK_SCAN_RAKING",
    "raking_memoize": "::cub::BlockScanAlgorithm::BLOCK_SCAN_RAKING_MEMOIZE",
    "warp_scans": "::cub::BlockScanAlgorithm::BLOCK_SCAN_WARP_SCANS",
}


CUB_BLOCK_REDUCE_ALGOS = {
    "raking_commutative_only": "::cub::BlockReduceAlgorithm::BLOCK_REDUCE_RAKING_COMMUTATIVE_ONLY",
    "raking": "::cub::BlockReduceAlgorithm::BLOCK_REDUCE_RAKING",
    "warp_reductions": "::cub::BlockReduceAlgorithm::BLOCK_REDUCE_WARP_REDUCTIONS",
}


class CudaSharedMemConfig(Enum):
    """
    CUDA shared memory configuration.  This is intended to mirror the C++
    equivalent `cudaSharedMemConfig` enum.
    """

    BankSizeDefault = 0
    BankSizeFourByte = 1
    BankSizeEightByte = 2

    def __str__(self):
        return f"cudaSharedMem{self.name}"


def make_binary_tempfile(content: bytes, suffix: str) -> BinaryIO:
    """
    Creates an unbuffered temporary binary file containing **content** and
    ending with **suffix**.  The caller is responsible for closing the file.

    :param content: Supplies the content to write to the temporary file.

    :param suffix: Supplies the suffix for the temporary file.

    :return: A binary file-like object representing the temporary file.
    """
    tmp = tempfile.NamedTemporaryFile(
        mode="w+b", suffix=suffix, buffering=0, delete=False
    )
    tmp.write(content)
    return tmp


def check_in(name, arg, set):
    if arg not in set:
        raise ValueError(f"{name} must be in {set} ; got {name} = {arg}")


def check_not_in(name, arg, set):
    if arg in set:
        raise ValueError(
            f"{name} must not be any of those value {set} ; got {name} = {arg}"
        )


def check_contains(set, key):
    if key not in set:
        raise ValueError(f"{key} must be in {set}")


def check_dim3(name, arg):
    if len(arg) != 3:
        raise ValueError(f"{name} should be a length-3 tuple ; got {name} = {arg}")


def find_unsigned(name, txt):
    regex = re.compile(f".global .align 4 .u32 {name} = ([0-9]*);", re.MULTILINE)
    found = regex.search(txt)
    if found is None:  # TODO: improve regex logic
        regex = re.compile(f".global .align 4 .u32 {name};", re.MULTILINE)
        found = regex.search(txt)
        if found is not None:
            return 0
        else:
            raise ValueError(f"{name} not found in text")
    else:
        return int(found.group(1))


def find_mangled_name(name, txt):
    regex = re.compile(f"[_a-zA-Z0-9]*{name}[_a-zA-Z0-9]*", re.MULTILINE)
    return regex.search(txt).group(0)


def find_dim2(name, txt):
    return (find_unsigned(f"{name}_x", txt), find_unsigned(f"{name}_y", txt))


def find_dim3(name, txt):
    return (
        find_unsigned(f"{name}_x", txt),
        find_unsigned(f"{name}_y", txt),
        find_unsigned(f"{name}_z", txt),
    )


def normalize_dim_param(dim: DimType) -> dim3:
    """
    Normalize the dim parameter to a `dim3` (x, y, z) instance.

    The logic for this routine is as follows:

    - If the dim is already a `dim3` instance, return it as is.
    - If the dim is a positive integer, return a 1D `dim3` instance with the
      integer value as the x-dimension.  If the dim is a negative integer,
      raise a ValueError.
    - If the dim is a tuple:
        - If the tuple has two elements, return a 2D `dim3` instance with the
          tuple values as the x and y dimensions.  If either value is
          negative, raise a ValueError.
        - If the tuple has three elements, return a 3D `dim3` instance with
          the tuple values as the x, y, and z dimensions.  If any value is
          negative, raise a ValueError.

    Args:
        dim: Supplies the dim parameter to normalize.

    Returns:
        The normalized dim parameter as a `dim3` instance.

    Raises:
        ValueError: If the dim is invalid.

    """
    if isinstance(dim, dim3):
        return dim

    if isinstance(dim, int):
        if dim < 0:
            msg = f"Dimension value must be non-negative, got {dim}"
            raise ValueError(msg)
        return dim3(dim, 1, 1)

    if isinstance(dim, tuple):
        if len(dim) == 2:
            x, y = dim
            z = 1
            if x < 0 or y < 0:
                msg = f"Dimension values must be non-negative, got {dim}"
                raise ValueError(msg)
            return dim3(x, y, z)
        elif len(dim) == 3:
            x, y, z = dim
            if x < 0 or y < 0 or z < 0:
                msg = f"Dimension values must be non-negative, got {dim}"
                raise ValueError(msg)
            return dim3(x, y, z)
        else:
            msg = f"Tuple dimension must have 2 or 3 elements, got {len(dim)}"
            raise ValueError(msg)

    raise ValueError(f"Unsupported dimension type: {type(dim)}")


def normalize_dtype_param(
    dtype: Union[str, type, "np.dtype", "numba.types.Type"],
) -> "numba.types.Type":
    """
    Normalize the dtype parameter to an appropriate Numba type.

    The logic for this routine is as follows:

    - If the dtype is already a numba type, return it as is.
    - If the dtype is a valid numpy dtype, convert it to the corresponding
      numba type.  Note that this applies to both `np.int32` and
      `np.dtype(np.int32)`.
    - If the dtype is a string:
        - If there's a period in the string, ensure it's because the string
          starts with "np." and is followed by a valid numpy dtype.  Otherwise,
          raise a ValueError in both cases: if there's a period belonging to
          something other than the leading "np.", or if the following numpy
          type isn't valid.
        - If there's no period, assume the type is referring to a numba type.
          If not, raise a ValueError.  If it is, return the corresponding numba
          type.

    Args:
        dtype: Supplies the dtype parameter to normalize.

    Returns:
        The normalized dtype parameter as a numba type.

    Raises:
        ValueError: If the dtype is invalid.
    """
    import numba
    import numpy as np

    # If dtype is already a numba type, return it as is.
    if isinstance(dtype, numba.types.Type):
        return dtype

    # Handle numpy dtype objects.
    if hasattr(np, "dtype") and isinstance(dtype, np.dtype):
        # Convert numpy dtype to numba type.
        try:
            return numba.from_dtype(dtype)
        except Exception as e:
            msg = f"Failed to convert numpy dtype {dtype} to numba type: {e}"
            raise ValueError(msg)

    is_type_like = (
        hasattr(np, "generic")
        and isinstance(dtype, type)
        and issubclass(dtype, np.generic)
    )

    # Handle numpy type objects (like np.int32).
    if is_type_like:
        try:
            return numba.from_dtype(np.dtype(dtype))
        except Exception as e:
            msg = f"Failed to convert numpy type {dtype} to numba type: {e}"
            raise ValueError(msg)

    # Handle string representations.
    if isinstance(dtype, str):
        if "." in dtype:
            # Check if string starts with "np." and is followed by a
            # valid numpy dtype.
            if not dtype.startswith("np."):
                msg = (
                    f"Invalid dtype string format: {dtype}. "
                    "String with period must start with 'np.' "
                    "and be followed by a valid numpy dtype."
                )
                raise ValueError(msg)

            # Extract the numpy type name.
            np_type_name = dtype[3:]

            # Check if it's a valid numpy type.
            if not hasattr(np, np_type_name):
                raise ValueError(f"Invalid numpy dtype: {np_type_name}")

            # Convert to numba type.
            try:
                np_type = getattr(np, np_type_name)
                return numba.from_dtype(np.dtype(np_type))
            except Exception as e:
                msg = f"Failed to convert numpy type {np_type_name} to numba type: {e}"
                raise ValueError(msg)
        else:
            # No period, assume it's a numba type.
            if not hasattr(numba, dtype):
                raise ValueError(f"Invalid numba type name: {dtype}")

            return getattr(numba, dtype)

    # If we get here, the dtype is not recognized.
    raise ValueError(f"Unrecognized dtype format: {dtype}")



# Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

"""ReverseIterator implementation."""

from __future__ import annotations

from textwrap import dedent

from .._bindings import Op, OpKind
from .._cpp_compile import compile_cpp_to_ltoir
from .._utils.protocols import get_size, is_device_array
from ._base import IteratorBase
from ._common import CUDA_PREAMBLE, ensure_iterator


class ReverseIterator(IteratorBase):
    """
    Iterator that reverses the direction of an underlying iterator.

    Advance with positive offset moves backward in the underlying iterator.
    """

    __slots__ = [
        "_underlying",
    ]

    def __init__(self, underlying):
        """
        Create a reverse iterator.

        Args:
            underlying: The underlying iterator or array to reverse
        """

        if is_device_array(underlying):
            # TODO: this is probably incorrect behaviour. In C++, initializing
            # with a pointer to the end of the array is left to be done explicitly
            # by the user.
            self._underlying = ensure_iterator(underlying) + (get_size(underlying) - 1)
        else:
            self._underlying = ensure_iterator(underlying)

        super().__init__(
            state_bytes=bytes(self._underlying.state),
            state_alignment=self._underlying.state_alignment,
            value_type=self._underlying.value_type,
        )

    def _make_advance_op(self) -> Op:
        """Provide Op for advance that negates offset direction."""
        child_op = self._underlying.get_advance_op()
        symbol = self._make_advance_symbol()

        source = dedent(f"""
            {CUDA_PREAMBLE}

            extern "C" __device__ void {child_op.name}(void* state, void* offset);

            extern "C" __device__ void {symbol}(void* state, void* offset) {{
                int64_t neg_offset = -static_cast<int64_t>(*static_cast<uint64_t*>(offset));
                {child_op.name}(state, &neg_offset);
            }}
        """).strip()

        ltoir = compile_cpp_to_ltoir(source)

        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[child_op.ltoir, *child_op.extra_ltoirs],
        )

    def _make_input_deref_op(self) -> Op | None:
        """Provide Op for input dereference that delegates to underlying."""
        child_op = self._underlying.get_input_deref_op()
        if child_op is None:
            return None

        symbol = self._make_input_deref_symbol()

        source = dedent(f"""
            {CUDA_PREAMBLE}

            extern "C" __device__ void {child_op.name}(void* state, void* result);

            extern "C" __device__ void {symbol}(void* state, void* result) {{
                {child_op.name}(state, result);
            }}
        """).strip()

        ltoir = compile_cpp_to_ltoir(source)

        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[child_op.ltoir, *child_op.extra_ltoirs],
        )

    def _make_output_deref_op(self) -> Op | None:
        """Provide Op for output dereference that delegates to underlying."""
        child_op = self._underlying.get_output_deref_op()
        if child_op is None:
            return None

        symbol = self._make_output_deref_symbol()

        source = dedent(f"""
            {CUDA_PREAMBLE}

            extern "C" __device__ void {child_op.name}(void* state, void* value);

            extern "C" __device__ void {symbol}(void* state, void* value) {{
                {child_op.name}(state, value);
            }}
        """).strip()

        ltoir = compile_cpp_to_ltoir(source)

        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[child_op.ltoir, *child_op.extra_ltoirs],
        )

    @property
    def children(self):
        return (self._underlying,)

    @property
    def kind(self):
        """Return a hashable kind for caching purposes."""
        return ("ReverseIterator", self._underlying.kind)

    def __add__(self, offset: int):
        return ReverseIterator(self._underlying + offset)



# Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception


import numba

from .._common import (
    make_binary_tempfile,
    normalize_dim_param,
    normalize_dtype_param,
)
from .._types import (
    Algorithm,
    Dependency,
    DependentArray,
    DependentPointer,
    Invocable,
    Pointer,
    TemplateParameter,
)

CUB_BLOCK_LOAD_ALGOS = {
    "direct": "::cub::BLOCK_LOAD_DIRECT",
    "striped": "::cub::BLOCK_LOAD_STRIPED",
    "vectorize": "::cub::BLOCK_LOAD_VECTORIZE",
    "transpose": "::cub::BLOCK_LOAD_TRANSPOSE",
    "warp_transpose": "::cub::BLOCK_LOAD_WARP_TRANSPOSE",
    "warp_transpose_timesliced": "::cub::BLOCK_LOAD_WARP_TRANSPOSE_TIMESLICED",
}

CUB_BLOCK_STORE_ALGOS = {
    "direct": "::cub::BLOCK_STORE_DIRECT",
    "striped": "::cub::BLOCK_STORE_STRIPED",
    "vectorize": "::cub::BLOCK_STORE_VECTORIZE",
    "transpose": "::cub::BLOCK_STORE_TRANSPOSE",
    "warp_transpose": "::cub::BLOCK_STORE_WARP_TRANSPOSE",
    "warp_transpose_timesliced": "::cub::BLOCK_STORE_WARP_TRANSPOSE_TIMESLICED",
}


def make_load(dtype, threads_per_block, items_per_thread=1, algorithm="direct"):
    """
    Creates a block-wide load primitive.

    Returns a callable object that can be linked to and invoked from
    device code. It can be invoked with the following signatures:

    - `(src: numba.types.Array, dest: numba.types.Array) -> None`:
      Each thread loads `items_per_thread` items from `src` into `dest`.
      `dest` must contain at least `items_per_thread` items.

    Different data movement strategies can be selected via the
    `algorithm` parameter:

    - `algorithm="direct"` (default): Reads blocked data directly.
    - `algorithm="striped"`: Reads striped data directly.
    - `algorithm="vectorize"`: Reads blocked data directly using
      CUDA's built-in vectorized loads as a coalescing optimization.
    - `algorithm="transpose"`: Reads striped data and then locally
      transposes it into a blocked arrangement.
    - `algorithm="warp_transpose"`: Reads warp-striped data and then
      locally transposes it into a blocked arrangement.
    - `algorithm="warp_transpose_timesliced"`: Reads warp-striped data
      and then locally transposes it into a blocked arrangement one
      warp at a time.

    For more details, read the corresponding CUB C++ documentation:
    https://nvidia.github.io/cccl/cub/api/classcub_1_1BlockLoad.html

    Args:
        dtype: Data type being loaded
        threads_per_block: Number of threads in a block. Can be an
            integer or a tuple of 2 or 3 integers.
        items_per_thread: The number of items each thread loads
        algorithm: The data movement algorithm to use

    Example:
        The code snippet below illustrates a striped load and store of
        128 integer items by 32 threads, with each thread handling
        4 integers.

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_block_load_store_api.py
            :language: python
            :dedent:
            :start-after: example-begin imports
            :end-before: example-end imports

        The following snippet shows how to invoke the returned
        ``block_load`` and ``block_store`` primitives:

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_block_load_store_api.py
            :language: python
            :dedent:
            :start-after: example-begin load_store
            :end-before: example-end load_store
    """
    dim = normalize_dim_param(threads_per_block)
    dtype = normalize_dtype_param(dtype)

    template = Algorithm(
        "BlockLoad",
        "Load",
        "block_load",
        ["cub/block/block_load.cuh"],
        [
            TemplateParameter("T"),
            TemplateParameter("BLOCK_DIM_X"),
            TemplateParameter("ITEMS_PER_THREAD"),
            TemplateParameter("ALGORITHM"),
            TemplateParameter("BLOCK_DIM_Y"),
            TemplateParameter("BLOCK_DIM_Z"),
        ],
        [
            [
                Pointer(numba.uint8),
                DependentPointer(Dependency("T")),
                DependentArray(Dependency("T"), Dependency("ITEMS_PER_THREAD")),
            ]
        ],
    )
    specialization = template.specialize(
        {
            "T": dtype,
            "BLOCK_DIM_X": dim[0],
            "ITEMS_PER_THREAD": items_per_thread,
            "ALGORITHM": CUB_BLOCK_LOAD_ALGOS[algorithm],
            "BLOCK_DIM_Y": dim[1],
            "BLOCK_DIM_Z": dim[2],
        }
    )
    return Invocable(
        temp_files=[
            make_binary_tempfile(ltoir, ".ltoir")
            for ltoir in specialization.get_lto_ir()
        ],
        temp_storage_bytes=specialization.temp_storage_bytes,
        temp_storage_alignment=specialization.temp_storage_alignment,
        algorithm=specialization,
    )


def make_store(dtype, threads_per_block, items_per_thread=1, algorithm="direct"):
    """
    Creates a block-wide store primitive.

    Returns a callable object that can be linked to and invoked from
    device code. It can be invoked with the following signatures:

    - `(dest: numba.types.Array, src: numba.types.Array) -> None`:
      Each thread stores `items_per_thread` items from `src` into `dest`.
      `src` must contain at least `items_per_thread` items.

    Different data movement strategies can be selected via the
    `algorithm` parameter:

    - `algorithm="direct"` (default): Writes blocked data directly.
    - `algorithm="striped"`: Writes striped data directly.
    - `algorithm="vectorize"`: Writes blocked data directly using
      CUDA's built-in vectorized stores as a coalescing optimization.
    - `algorithm="transpose"`: Locally transposes blocked data into a
      striped arrangement before writing to memory.
    - `algorithm="warp_transpose"`: Locally transposes blocked data
      into a warp-striped arrangement before writing to memory.
    - `algorithm="warp_transpose_timesliced"`: Locally transposes
      blocked data into a warp-striped arrangement before writing to
      memory. To reduce shared memory requirements, only one warp's
      worth of shared memory is provisioned and time-sliced across warps.

    For more details, read the corresponding CUB C++ documentation:
    https://nvidia.github.io/cccl/cub/api/classcub_1_1BlockStore.html

    Args:
        dtype: Data type being stored
        threads_per_block: Number of threads in a block. Can be an
            integer or a tuple of 2 or 3 integers.
        items_per_thread: The number of items each thread stores
        algorithm: The data movement algorithm to use

    Example:
        The code snippet below illustrates a striped load and store of
        128 integer items by 32 threads, with each thread handling
        4 integers.

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_block_load_store_api.py
            :language: python
            :dedent:
            :start-after: example-begin imports
            :end-before: example-end imports

        The following snippet shows how to invoke the returned
        ``block_load`` and ``block_store`` primitives:

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_block_load_store_api.py
            :language: python
            :dedent:
            :start-after: example-begin load_store
            :end-before: example-end load_store
    """
    dim = normalize_dim_param(threads_per_block)
    dtype = normalize_dtype_param(dtype)

    template = Algorithm(
        "BlockStore",
        "Store",
        "block_store",
        ["cub/block/block_store.cuh"],
        [
            TemplateParameter("T"),
            TemplateParameter("BLOCK_DIM_X"),
            TemplateParameter("ITEMS_PER_THREAD"),
            TemplateParameter("ALGORITHM"),
            TemplateParameter("BLOCK_DIM_Y"),
            TemplateParameter("BLOCK_DIM_Z"),
        ],
        [
            [
                Pointer(numba.uint8),
                DependentPointer(Dependency("T")),
                DependentArray(Dependency("T"), Dependency("ITEMS_PER_THREAD")),
            ]
        ],
    )
    specialization = template.specialize(
        {
            "T": dtype,
            "BLOCK_DIM_X": dim[0],
            "ITEMS_PER_THREAD": items_per_thread,
            "ALGORITHM": CUB_BLOCK_STORE_ALGOS[algorithm],
            "BLOCK_DIM_Y": dim[1],
            "BLOCK_DIM_Z": dim[2],
        }
    )
    return Invocable(
        temp_files=[
            make_binary_tempfile(ltoir, ".ltoir")
            for ltoir in specialization.get_lto_ir()
        ],
        temp_storage_bytes=specialization.temp_storage_bytes,
        temp_storage_alignment=specialization.temp_storage_alignment,
        algorithm=specialization,
    )



# Copyright (c) 2024-2025, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from ._block_exchange import (
    BlockExchangeType,
    make_exchange,
)
from ._block_load_store import make_load, make_store
from ._block_merge_sort import make_merge_sort_keys
from ._block_radix_sort import (
    make_radix_sort_keys,
    make_radix_sort_keys_descending,
)
from ._block_reduce import make_reduce, make_sum
from ._block_scan import (
    make_exclusive_scan,
    make_exclusive_sum,
    make_inclusive_scan,
    make_inclusive_sum,
    make_scan,
)

__all__ = [
    "BlockExchangeType",
    "make_exchange",
    "make_exclusive_scan",
    "make_exclusive_sum",
    "make_inclusive_scan",
    "make_inclusive_sum",
    "make_load",
    "make_merge_sort_keys",
    "make_radix_sort_keys",
    "make_radix_sort_keys_descending",
    "make_reduce",
    "make_scan",
    "make_store",
    "make_sum",
]



# Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

"""PointerIterator implementation - simple bidirectional iterator for device arrays."""

from __future__ import annotations

import ctypes
import sys
from textwrap import dedent

from .._bindings import Op, OpKind
from .._cpp_compile import compile_cpp_to_ltoir, cpp_type_from_descriptor
from .._utils.protocols import get_data_pointer, get_dtype
from ..types import from_numpy_dtype
from ._base import IteratorBase
from ._common import CUDA_PREAMBLE


class PointerIterator(IteratorBase):
    """
    Simple iterator wrapping a device array pointer.

    Supports both input (reading) and output (writing) operations.
    Handles both scalar types (using typed C++ code) and struct types
    (using byte-level memcpy).
    """

    def __init__(self, array):
        """
        Create a pointer iterator from a device array.

        Args:
            array: Device array with __cuda_array_interface__
        """
        # Get pointer and dtype from array
        ptr = get_data_pointer(array)
        dtype = get_dtype(array)
        value_type = from_numpy_dtype(dtype)

        # State is just the pointer
        state_bytes = ctypes.c_void_p(ptr)
        state_bytes_buffer = (ctypes.c_char * 8)()
        ctypes.memmove(state_bytes_buffer, ctypes.byref(state_bytes), 8)
        state_bytes = bytes(state_bytes_buffer)

        self._cpp_type = cpp_type_from_descriptor(value_type)  # None for struct types
        self._element_size = value_type.info.size
        self._array = array  # Keep reference to prevent GC

        super().__init__(
            state_bytes=state_bytes,
            state_alignment=8,  # pointer alignment
            value_type=value_type,
        )

    @property
    def array(self):
        return self._array

    def _make_advance_op(self) -> Op:
        symbol = self._make_advance_symbol()

        if self._cpp_type:
            # Scalar type - use typed pointer arithmetic
            source = dedent(f"""
                {CUDA_PREAMBLE}

                extern "C" __device__ void {symbol}(void* state, void* offset) {{
                    auto* ptr_state = static_cast<{self._cpp_type}**>(state);
                    auto dist = *static_cast<int64_t*>(offset);
                    *ptr_state += dist;
                }}
            """).strip()
        else:
            # Struct type - use byte-level pointer arithmetic
            source = dedent(f"""
                {CUDA_PREAMBLE}

                extern "C" __device__ void {symbol}(void* state, void* offset) {{
                    auto* ptr_state = static_cast<char**>(state);
                    auto dist = *static_cast<int64_t*>(offset);
                    *ptr_state += dist * {self._element_size};
                }}
            """).strip()

        ltoir = compile_cpp_to_ltoir(source)
        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[],
        )

    def _make_input_deref_op(self) -> Op | None:
        symbol = self._make_input_deref_symbol()

        if self._cpp_type:
            # Scalar type - use typed dereference
            source = dedent(f"""
                {CUDA_PREAMBLE}

                extern "C" __device__ void {symbol}(void* state, void* result) {{
                    auto* ptr_state = static_cast<{self._cpp_type}**>(state);
                    *static_cast<{self._cpp_type}*>(result) = **ptr_state;
                }}
            """).strip()
        else:
            # Struct type - use memcpy
            source = dedent(f"""
                {CUDA_PREAMBLE}

                extern "C" __device__ void {symbol}(void* state, void* result) {{
                    auto* ptr_state = static_cast<char**>(state);
                    memcpy(result, *ptr_state, {self._element_size});
                }}
            """).strip()

        ltoir = compile_cpp_to_ltoir(source)
        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[],
        )

    def _make_output_deref_op(self) -> Op | None:
        symbol = self._make_output_deref_symbol()

        if self._cpp_type:
            # Scalar type - use typed dereference
            source = dedent(f"""
                {CUDA_PREAMBLE}

                extern "C" __device__ void {symbol}(void* state, void* value) {{
                    auto* ptr_state = static_cast<{self._cpp_type}**>(state);
                    **ptr_state = *static_cast<{self._cpp_type}*>(value);
                }}
            """).strip()
        else:
            # Struct type - use memcpy
            source = dedent(f"""
                {CUDA_PREAMBLE}

                extern "C" __device__ void {symbol}(void* state, void* value) {{
                    auto* ptr_state = static_cast<char**>(state);
                    memcpy(*ptr_state, value, {self._element_size});
                }}
            """).strip()

        ltoir = compile_cpp_to_ltoir(source)
        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[],
        )

    def __add__(self, offset: int):
        dtype = get_dtype(self._array)
        offset_ptr = self._current_pointer() + offset * dtype.itemsize
        return self._clone_with_pointer(offset_ptr)

    def _current_pointer(self) -> int:
        return int.from_bytes(self._state_bytes, sys.byteorder, signed=False)

    def _clone_with_pointer(self, pointer_value: int):
        """Clone this iterator with a different pointer value."""
        clone = PointerIterator(self._array)
        state_bytes_buffer = (ctypes.c_char * 8)()
        ptr_obj = ctypes.c_void_p(pointer_value)
        ctypes.memmove(state_bytes_buffer, ctypes.byref(ptr_obj), 8)
        clone._state_bytes = bytes(state_bytes_buffer)
        clone._uid_cached = None
        return clone

    @property
    def kind(self):
        """
        Return a hashable kind for caching purposes.

        Include _cpp_type and _element_size since they affect generated code.
        Different code paths are taken for scalar vs struct types.
        """
        return (
            type(self).__name__,
            self._value_type,
            self._cpp_type,
            self._element_size,
        )



# Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import numba

from .._common import make_binary_tempfile
from .._types import (
    Algorithm,
    Dependency,
    DependentPythonOperator,
    DependentReference,
    Invocable,
    Pointer,
    TemplateParameter,
    numba_type_to_wrapper,
)


def make_reduce(dtype, binary_op, threads_in_warp=32, methods=None):
    """
    Creates a warp-wide reduction primitive for lane :sub:`0` using the
    specified binary reduction functor. Each thread contributes one input
    element.

    Warning:
        The return value is undefined in threads other than thread
        :sub:`0`.

    Example:
        The code snippet below illustrates a max reduction of 32 integer
        items that are partitioned across a warp of threads.

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_warp_reduce_api.py
            :language: python
            :dedent:
            :start-after: example-begin imports
            :end-before: example-end imports

        The following snippet shows how to invoke the returned
        ``warp_reduce`` primitive:

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_warp_reduce_api.py
            :language: python
            :dedent:
            :start-after: example-begin reduce
            :end-before: example-end reduce

        Suppose the set of inputs across the warp of threads is
        ``{ 0, 1, 2, 3, ..., 31 }``.
        The corresponding output in lane :sub:`0` will be ``{ 31 }``.

    Args:
        dtype: Data type being reduced
        threads_in_warp: The number of threads in a warp
        binary_op: Binary reduction function

    Returns:
        A callable object that can be linked to and invoked from a CUDA
        kernel
    """
    template = Algorithm(
        "WarpReduce",
        "Reduce",
        "warp_reduce",
        ["cub/warp/warp_reduce.cuh"],
        [TemplateParameter("T"), TemplateParameter("VIRTUAL_WARP_THREADS")],
        [
            [
                Pointer(numba.uint8),
                DependentReference(Dependency("T")),
                DependentPythonOperator(
                    Dependency("T"),
                    [Dependency("T"), Dependency("T")],
                    Dependency("Op"),
                ),
                DependentReference(Dependency("T"), True),
            ]
        ],
        type_definitions=[numba_type_to_wrapper(dtype, methods=methods)],
    )
    specialization = template.specialize(
        {"T": dtype, "VIRTUAL_WARP_THREADS": threads_in_warp, "Op": binary_op}
    )

    return Invocable(
        temp_files=[
            make_binary_tempfile(ltoir, ".ltoir")
            for ltoir in specialization.get_lto_ir(threads=threads_in_warp)
        ],
        temp_storage_bytes=specialization.temp_storage_bytes,
        temp_storage_alignment=specialization.temp_storage_alignment,
        algorithm=specialization,
    )


def make_sum(dtype, threads_in_warp=32):
    """
    Creates a warp-wide reduction primitive for lane :sub:`0` using
    addition (+) as the reduction operator. Each thread contributes one
    input element.

    Warning:
        The return value is undefined in threads other than thread
        :sub:`0`.

    Example:
        The code snippet below illustrates a reduction of 32 integer
        items that are partitioned across a warp of threads.

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_warp_reduce_api.py
            :language: python
            :dedent:
            :start-after: example-begin imports
            :end-before: example-end imports

        The following snippet shows how to invoke the returned
        ``warp_sum`` primitive:

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_warp_reduce_api.py
            :language: python
            :dedent:
            :start-after: example-begin sum
            :end-before: example-end sum

        Suppose the set of inputs across the warp of threads is
        ``{ 1, 1, 1, 1, ..., 1 }``.
        The corresponding output in lane :sub:`0` will be ``{ 32 }``.

    Args:
        dtype: Data type being reduced
        threads_in_warp: The number of threads in a warp

    Returns:
        A callable object that can be linked to and invoked from a CUDA
        kernel
    """
    template = Algorithm(
        "WarpReduce",
        "Sum",
        "warp_reduce",
        ["cub/warp/warp_reduce.cuh"],
        [TemplateParameter("T"), TemplateParameter("VIRTUAL_WARP_THREADS")],
        [
            [
                Pointer(numba.uint8),
                DependentReference(Dependency("T")),
                DependentReference(Dependency("T"), True),
            ]
        ],
    )
    specialization = template.specialize(
        {"T": dtype, "VIRTUAL_WARP_THREADS": threads_in_warp}
    )
    return Invocable(
        temp_files=[
            make_binary_tempfile(ltoir, ".ltoir")
            for ltoir in specialization.get_lto_ir(threads=threads_in_warp)
        ],
        temp_storage_bytes=specialization.temp_storage_bytes,
        temp_storage_alignment=specialization.temp_storage_alignment,
        algorithm=specialization,
    )



"""
CUDA Core Library (CCCL) Python Package
"""

import importlib.metadata

try:
    __version__ = importlib.metadata.version("cuda-cccl")
except Exception:
    __version__ = "0.0.0"

from .headers.include_paths import get_include_paths

# cuda.bindings is required, but instead of being listed as a required dependency,
# it is installed via an extra (e.g., [cu12] or [cu13]).
#
# One of the first things we should do is check that it is available, and raise
# a helpful error message if it is not.
try:
    import cuda.bindings as _cuda_bindings  # type: ignore
except ImportError:
    raise ImportError(
        "cuda.bindings is not installed. Please install the appropriate extra cuda-cccl[cu12] or cuda-cccl[cu13]."
    ) from None
del _cuda_bindings

__all__ = ["get_include_paths", "__version__"]



from __future__ import annotations

import functools
import weakref
from types import SimpleNamespace
from typing import Optional

from cuda.bindings import driver, runtime

try:
    from cuda.core import Device
    from cuda.core._utils.cuda_utils import handle_return
except ImportError:
    from cuda.core.experimental import Device
    from cuda.core.experimental._utils.cuda_utils import handle_return

from ..typing import StreamLike


@functools.cache
def _set_default_mempool_threshold(device_id: int):
    """
    Set the release threshold for the default memory pool on this device,
    if we haven't already done so. This prevents the driver from attempting
    to shrink the pool after every sync, which can be slow.
    """
    default_pool = handle_return(driver.cuDeviceGetDefaultMemPool(device_id))
    threshold = handle_return(
        driver.cuMemPoolGetAttribute(
            default_pool, driver.CUmemPool_attribute.CU_MEMPOOL_ATTR_RELEASE_THRESHOLD
        )
    )
    if int(threshold) == 0:
        handle_return(
            driver.cuMemPoolSetAttribute(
                default_pool,
                driver.CUmemPool_attribute.CU_MEMPOOL_ATTR_RELEASE_THRESHOLD,
                driver.cuuint64_t(0xFFFFFFFFFFFFFFFF),
            )
        )


def _finalize_buffer(ptr: int, stream_handle: Optional[int] = None):
    """Cleanup function for weakref finalizer."""
    if ptr != 0:
        try:
            handle_return(runtime.cudaFreeAsync(ptr, stream_handle))
        except Exception as e:
            # Don't raise in finalizer, just print warning
            print(f"Warning: Failed to free CUDA memory: {e}")


class TempStorageBuffer:
    """
    Simple wrapper type around the memory allocation used for temporary storage,
    exposing __cuda_array_interface__ and some other attributes for fast access.

    This implementation uses cuda.bindings.runtime.cudaMallocAsync and
    cudaFreeAsync for allocation and deallocation.
    """

    def __init__(self, size: int, stream: Optional[StreamLike] = None):
        # Get the current device
        dev = Device()

        stream_handle = stream.__cuda_stream__()[1] if stream is not None else None

        # Set the release threshold for the default memory pool on this device
        _set_default_mempool_threshold(dev.device_id)

        # Allocate memory using cudaMallocAsync
        device_ptr_int = handle_return(runtime.cudaMallocAsync(size, stream_handle))
        self._ptr = int(device_ptr_int)
        self._stream_handle = stream_handle
        self._size = size

        # attributes for fast path access in protocols.py
        self.nbytes = size
        self.data = SimpleNamespace(ptr=self._ptr)

        # Set up weakref finalizer for cleanup
        self._finalizer = weakref.finalize(
            self, _finalize_buffer, self._ptr, self._stream_handle
        )

    @property
    def __cuda_array_interface__(self):
        return {
            "data": (self._ptr, False),
            "shape": (self._size,),
            "strides": (1,),
            "typestr": "|u1",
            "version": 3,
        }



# Copyright (c) 2024-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from __future__ import annotations

from typing import Callable

import numpy as np

from .. import _bindings
from .. import _cccl_interop as cccl
from .._caching import cache_with_registered_key_functions
from .._cccl_interop import (
    call_build,
    get_value_type,
    set_cccl_iterator_state,
    to_cccl_value_state,
)
from .._utils.protocols import get_data_pointer, validate_and_get_stream
from .._utils.temp_storage_buffer import TempStorageBuffer
from ..determinism import Determinism
from ..op import OpAdapter, make_op_adapter
from ..typing import DeviceArrayLike, GpuStruct, IteratorT, Operator


class _Reduce:
    __slots__ = [
        "d_in_cccl",
        "d_out_cccl",
        "h_init_cccl",
        "op_cccl",
        "build_result",
        "device_reduce_fn",
    ]

    # TODO: constructor shouldn't require concrete `d_in`, `d_out`:
    def __init__(
        self,
        d_in: DeviceArrayLike | IteratorT,
        d_out: DeviceArrayLike | IteratorT,
        op: OpAdapter,
        h_init: np.ndarray | GpuStruct,
        determinism: Determinism,
    ):
        self.d_in_cccl = cccl.to_cccl_input_iter(d_in)
        self.d_out_cccl = cccl.to_cccl_output_iter(d_out)
        self.h_init_cccl = cccl.to_cccl_value(h_init)

        # Compile the op with value types
        value_type = get_value_type(h_init)
        self.op_cccl = op.compile((value_type, value_type), value_type)

        self.build_result = call_build(
            _bindings.DeviceReduceBuildResult,
            self.d_in_cccl,
            self.d_out_cccl,
            self.op_cccl,
            self.h_init_cccl,
            determinism,
        )

        match determinism:
            case Determinism.RUN_TO_RUN:
                self.device_reduce_fn = self.build_result.compute
            case Determinism.NOT_GUARANTEED:
                self.device_reduce_fn = self.build_result.compute_nondeterministic
            case _:
                raise ValueError(f"Invalid determinism: {determinism}")

    def __call__(
        self,
        temp_storage,
        d_in,
        d_out,
        op: Callable | OpAdapter,
        num_items: int,
        h_init: np.ndarray | GpuStruct,
        stream=None,
    ):
        set_cccl_iterator_state(self.d_in_cccl, d_in)
        set_cccl_iterator_state(self.d_out_cccl, d_out)

        # Update op state for stateful ops
        op_adapter = make_op_adapter(op)
        self.op_cccl.state = op_adapter.get_state()

        self.h_init_cccl.state = to_cccl_value_state(h_init)

        stream_handle = validate_and_get_stream(stream)

        if temp_storage is None:
            temp_storage_bytes = 0
            d_temp_storage = 0
        else:
            temp_storage_bytes = temp_storage.nbytes
            d_temp_storage = get_data_pointer(temp_storage)

        temp_storage_bytes = self.device_reduce_fn(
            d_temp_storage,
            temp_storage_bytes,
            self.d_in_cccl,
            self.d_out_cccl,
            num_items,
            self.op_cccl,
            self.h_init_cccl,
            stream_handle,
        )
        return temp_storage_bytes


@cache_with_registered_key_functions
def make_reduce_into(
    d_in: DeviceArrayLike | IteratorT,
    d_out: DeviceArrayLike | IteratorT,
    op: Operator,
    h_init: np.ndarray | GpuStruct,
    **kwargs,
):
    """Computes a device-wide reduction using the specified binary ``op`` and initial value ``init``.

    Example:
        Below, ``make_reduce_into`` is used to create a reduction object that can be reused.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/reduction/reduce_object.py
            :language: python
            :start-after: # example-begin


    Args:
        d_in: Device array or iterator containing the input sequence of data items
        d_out: Device array (of size 1) that will store the result of the reduction
        op: Binary operator to apply.
            The signature is ``(T, T) -> T``, where ``T`` is
            the data type of the initial value ``h_init``.
        init: Numpy array storing initial value of the reduction

    Returns:
        A callable object that can be used to perform the reduction
    """
    op_adapter = make_op_adapter(op)
    return _Reduce(
        d_in,
        d_out,
        op_adapter,
        h_init,
        kwargs.get("determinism", Determinism.RUN_TO_RUN),
    )


def reduce_into(
    d_in: DeviceArrayLike | IteratorT,
    d_out: DeviceArrayLike | IteratorT,
    op: Operator,
    num_items: int,
    h_init: np.ndarray | GpuStruct,
    stream=None,
    **kwargs,
):
    """
    Performs device-wide reduction.

    This function automatically handles temporary storage allocation and execution.

    Example:
        Below, ``reduce_into`` is used to compute the sum of a sequence of integers.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/reduction/sum_reduction.py
            :language: python
            :start-after: # example-begin

    Args:
        d_in: Device array or iterator containing the input sequence of data items
        d_out: Device array to store the result of the reduction
        op: Binary operator to apply.
            The signature is ``(T, T) -> T``, where ``T`` is
            the data type of the initial value ``h_init``.
        num_items: Number of items to reduce
        h_init: Initial value for the reduction
        stream: CUDA stream for the operation (optional)
    """
    reducer = make_reduce_into(d_in, d_out, op, h_init, **kwargs)
    tmp_storage_bytes = reducer(None, d_in, d_out, op, num_items, h_init, stream)
    tmp_storage = TempStorageBuffer(tmp_storage_bytes, stream)
    reducer(tmp_storage, d_in, d_out, op, num_items, h_init, stream)



# Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

"""ZipIterator implementation."""

from __future__ import annotations

from textwrap import dedent

from .._bindings import Op, OpKind
from .._cpp_compile import compile_cpp_to_ltoir
from ..types import struct
from ._base import IteratorBase, compose_iterator_states
from ._common import CUDA_PREAMBLE, ensure_iterator


class ZipIterator(IteratorBase):
    """
    Iterator that zips multiple iterators together.

    At each position, yields a tuple of values from all underlying iterators.

    Similar to `thrust::zip_iterator <https://nvidia.github.io/cccl/thrust/api/classthrust_1_1zip__iterator.html>`_.

    Example:
        The code snippet below demonstrates how to zip together an array and a
        :class:`CountingIterator <cuda.compute.iterators.CountingIterator>` to
        find the index of the maximum value of the array.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/iterator/zip_iterator_counting.py
            :language: python
            :start-after: # example-begin
    """

    __slots__ = [
        "_iterators",
        "_field_names",
        "_value_offsets",
        "_state_offsets",
        "_advance_result",
        "_input_deref_result",
        "_output_deref_result",
    ]

    def __init__(self, *args):
        """
        Create a zip iterator.

        Args:
            *args: Iterators or arrays to zip together. Can be:
                   - Multiple iterators/arrays: ZipIterator(it1, it2, it3)
                   - A single sequence of iterators: ZipIterator([it1, it2, it3])
        """
        # Handle both ZipIterator(it1, it2) and ZipIterator([it1, it2])
        if len(args) == 1 and isinstance(args[0], (list, tuple)):
            iterators = args[0]
        else:
            iterators = args

        if len(iterators) < 1:
            raise ValueError("ZipIterator requires at least one iterator")

        # Wrap arrays in PointerIterator
        iterators = [ensure_iterator(it) for it in iterators]

        self._iterators = list(iterators)

        # Compose states from all iterators
        self._state_bytes, self._state_alignment, self._state_offsets = (
            compose_iterator_states(self._iterators)
        )

        # Build combined value type (struct layout)
        self._field_names = [f"field_{i}" for i in range(len(self._iterators))]
        fields = {
            name: it.value_type for name, it in zip(self._field_names, self._iterators)
        }
        self._value_type = struct(fields, name=f"Zip{len(iterators)}")
        self._value_offsets = [
            self._value_type.dtype.fields[name][1] for name in self._field_names
        ]

        super().__init__(
            state_bytes=self._state_bytes,
            state_alignment=self._state_alignment,
            value_type=self._value_type,
        )

    def _make_advance_op(self) -> Op:
        """Provide Op for advance that calls all child iterator advances."""
        child_ops = [it.get_advance_op() for it in self._iterators]
        symbol = self._make_advance_symbol()

        externs = "\n".join(
            f'extern "C" __device__ void {op.name}(void* state, void* offset);'
            for op in child_ops
        )

        calls = "\n        ".join(
            f"{op.name}(static_cast<char*>(state) + {offset}, offset);"
            for op, offset in zip(child_ops, self._state_offsets)
        )

        source = dedent(f"""
            {CUDA_PREAMBLE}

            {externs}

            extern "C" __device__ void {symbol}(void* state, void* offset) {{
                {calls}
            }}
        """).strip()

        ltoir = compile_cpp_to_ltoir(source)

        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[
                ltoir for op in child_ops for ltoir in [op.ltoir, *op.extra_ltoirs]
            ],
        )

    def _make_input_deref_op(self) -> Op | None:
        """Provide Op for input deref that calls all child iterator input derefs."""
        child_ops = [it.get_input_deref_op() for it in self._iterators]
        if not all(op is not None for op in child_ops):
            return None

        symbol = self._make_input_deref_symbol()

        externs = "\n".join(
            f'extern "C" __device__ void {op.name}(void* state, void* result);'
            for op in child_ops
        )

        calls = "\n        ".join(
            f"{op.name}(static_cast<char*>(state) + {state_off}, "
            f"static_cast<char*>(result) + {val_off});"
            for op, state_off, val_off in zip(
                child_ops, self._state_offsets, self._value_offsets
            )
        )

        source = dedent(f"""
            {CUDA_PREAMBLE}

            {externs}

            extern "C" __device__ void {symbol}(void* state, void* result) {{
                {calls}
            }}
        """).strip()

        ltoir = compile_cpp_to_ltoir(source)

        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[
                ltoir for op in child_ops for ltoir in [op.ltoir, *op.extra_ltoirs]
            ],
        )

    def _make_output_deref_op(self) -> Op | None:
        """Provide Op for output deref that calls all child iterator output derefs."""
        child_ops = [it.get_output_deref_op() for it in self._iterators]
        if not all(op is not None for op in child_ops):
            return None

        symbol = self._make_output_deref_symbol()

        externs = "\n".join(
            f'extern "C" __device__ void {op.name}(void* state, void* value);'
            for op in child_ops
        )

        calls = "\n        ".join(
            f"{op.name}(static_cast<char*>(state) + {state_off}, "
            f"static_cast<char*>(value) + {val_off});"
            for op, state_off, val_off in zip(
                child_ops, self._state_offsets, self._value_offsets
            )
        )

        source = dedent(f"""
            {CUDA_PREAMBLE}

            {externs}

            extern "C" __device__ void {symbol}(void* state, void* value) {{
                {calls}
            }}
        """).strip()

        ltoir = compile_cpp_to_ltoir(source)

        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[
                ltoir for op in child_ops for ltoir in [op.ltoir, *op.extra_ltoirs]
            ],
        )

    @property
    def children(self):
        return tuple(self._iterators)

    def __add__(self, offset: int) -> "ZipIterator":
        """Advance all child iterators by offset."""
        advanced_iterators = [it + offset for it in self._iterators]  # type: ignore[operator]
        return ZipIterator(*advanced_iterators)

    @property
    def kind(self):
        """Return a hashable kind for caching purposes."""
        return ("ZipIterator", tuple(it.kind for it in self._iterators))



# Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import ctypes
from enum import IntEnum
from typing import Any, Optional

from typing_extensions import Buffer

class OpKind(IntEnum):
    _value_: int
    STATELESS = ...
    STATEFUL = ...
    PLUS = ...
    MINUS = ...
    MULTIPLIES = ...
    DIVIDES = ...
    MODULUS = ...
    EQUAL_TO = ...
    NOT_EQUAL_TO = ...
    GREATER = ...
    LESS = ...
    GREATER_EQUAL = ...
    LESS_EQUAL = ...
    LOGICAL_AND = ...
    LOGICAL_OR = ...
    LOGICAL_NOT = ...
    BIT_AND = ...
    BIT_OR = ...
    BIT_XOR = ...
    BIT_NOT = ...
    IDENTITY = ...
    NEGATE = ...
    MINIMUM = ...
    MAXIMUM = ...

class TypeEnum(IntEnum):
    _value_: int
    INT8 = ...
    INT16 = ...
    INT32 = ...
    INT64 = ...
    UINT8 = ...
    UINT16 = ...
    UINT32 = ...
    UINT64 = ...
    FLOAT16 = ...
    FLOAT32 = ...
    FLOAT64 = ...
    STORAGE = ...
    BOOLEAN = ...

class IteratorKind(IntEnum):
    _value_: int
    POINTER = ...
    ITERATOR = ...

class SortOrder(IntEnum):
    _value_: int
    ASCENDING = ...
    DESCENDING = ...

class InitKind(IntEnum):
    _value_: int
    NO_INIT = ...
    FUTURE_VALUE_INIT = ...
    VALUE_INIT = ...

class Determinism(IntEnum):
    _value_: int
    NOT_GUARANTEED = ...
    RUN_TO_RUN = ...
    GPU_TO_GPU = ...

class BinarySearchMode(IntEnum):
    _value_: int
    LOWER_BOUND = ...
    UPPER_BOUND = ...

class Op:
    def __init__(
        self,
        name: Optional[str] = ...,
        operator_type: OpKind = ...,
        ltoir=None,
        state=None,
        state_alignment: int = 1,
        extra_ltoirs=None,
    ): ...
    @property
    def state(self) -> bytes: ...
    @state.setter
    def state(self, new_value: bytes) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def ltoir(self) -> bytes: ...
    @property
    def state_alignment(self) -> int: ...
    @property
    def state_typenum(self) -> int: ...
    def as_bytes(self) -> bytes: ...
    @property
    def extra_ltoirs(self) -> list: ...

class TypeInfo:
    def __init__(self, size: int, alignment: int, type_enum: TypeEnum): ...
    @property
    def size(self) -> int: ...
    @property
    def alignment(self) -> int: ...
    @property
    def typenum(self) -> int: ...
    def as_bytes(self) -> bytes: ...

class Value:
    def __init__(self, type: TypeInfo, state: memoryview): ...
    @property
    def type(self) -> TypeInfo: ...
    @property
    def state(self) -> memoryview: ...
    @state.setter
    def state(self, new_value: memoryview) -> None: ...
    def as_bytes(self) -> bytes: ...

class Pointer:
    def __init__(self, arg): ...

def make_pointer_object(ptr: int | ctypes.c_void_p, owner: Any) -> Pointer: ...

class IteratorState(Buffer):
    def __init__(self, arg): ...
    @property
    def size(self) -> int: ...

class Iterator:
    def __init__(
        self,
        alignment: int,
        iterator_type: IteratorKind,
        advance_fn: Op,
        dereference_fn: Op,
        value_type: TypeInfo,
        state=None,
        host_advance_fn=None,
    ):
        pass

    @property
    def advance_op(self) -> Op: ...
    @property
    def dereference_op(self) -> Op: ...
    @property
    def state(self): ...
    @state.setter
    def state(self, value) -> None: ...
    @property
    def type(self) -> IteratorKind: ...
    @property
    def value_type(self) -> TypeInfo: ...
    def as_bytes(self) -> bytes: ...
    def is_kind_pointer(self) -> bool: ...
    def is_kind_iterator(self) -> bool: ...
    @property
    def host_advance_fn(self): ...
    @host_advance_fn.setter
    def host_advance_fn(self, value) -> None: ...

class CommonData:
    def __init__(
        self,
        cc_major: int,
        cc_minor: int,
        cub_path: str,
        thrust_path: str,
        libcudacxx_path: str,
        ctk_path: str,
    ): ...
    @property
    def compute_capability(self) -> tuple[int, int]: ...
    @property
    def cub_path(self) -> str: ...
    @property
    def thrust_path(self) -> str: ...
    @property
    def libcudacxx_path(self) -> str: ...
    @property
    def ctk_path(self) -> str: ...

# ------------
# DeviceReduce
# ------------

class DeviceReduceBuildResult:
    def __init__(
        self,
        d_in: Iterator,
        d_out: Iterator,
        binary_op: Op,
        h_init: Value,
        determinism: Determinism,
        info: CommonData,
    ): ...
    def compute(
        self,
        temp_storage_ptr: int | None,
        temp_storage_nbytes: int,
        d_in: Iterator,
        d_out: Iterator,
        num_items: int,
        binary_op: Op,
        h_init: Value,
        stream,
    ) -> int: ...
    def compute_nondeterministic(
        self,
        temp_storage_ptr: int | None,
        temp_storage_nbytes: int,
        d_in: Iterator,
        d_out: Iterator,
        num_items: int,
        binary_op: Op,
        h_init: Value,
        stream,
    ) -> int: ...

# ----------
# DeviceScan
# ----------

class DeviceScanBuildResult:
    def __init__(
        self,
        d_in: Iterator,
        d_out: Iterator,
        binary_op: Op,
        init_type: TypeInfo,
        force_inclusive: bool,
        init_kind: InitKind,
        info: CommonData,
    ): ...
    def compute_inclusive(
        self,
        temp_storage_ptr: int | None,
        temp_storage_nbytes: int,
        d_in: Iterator,
        d_out: Iterator,
        num_items: int,
        binary_op: Op,
        h_init: Value,
        stream,
    ) -> int: ...
    def compute_exclusive(
        self,
        temp_storage_ptr: int | None,
        temp_storage_nbytes: int,
        d_in: Iterator,
        d_out: Iterator,
        num_items: int,
        binary_op: Op,
        h_init: Value,
        stream,
    ) -> int: ...
    def compute_inclusive_future_value(
        self,
        temp_storage_ptr: int | None,
        temp_storage_nbytes: int,
        d_in: Iterator,
        d_out: Iterator,
        num_items: int,
        binary_op: Op,
        h_init: Iterator,
        stream,
    ) -> int: ...
    def compute_exclusive_future_value(
        self,
        temp_storage_ptr: int | None,
        temp_storage_nbytes: int,
        d_in: Iterator,
        d_out: Iterator,
        num_items: int,
        binary_op: Op,
        h_init: Iterator,
        stream,
    ) -> int: ...
    def compute_inclusive_no_init(
        self,
        temp_storage_ptr: int | None,
        temp_storage_nbytes: int,
        d_in: Iterator,
        d_out: Iterator,
        num_items: int,
        binary_op: Op,
        h_init: None,
        stream,
    ) -> int: ...

# ---------------------
# DeviceSegmentedReduce
# ---------------------

class DeviceSegmentedReduceBuildResult:
    def __init__(
        self,
        d_in: Iterator,
        d_out: Iterator,
        start_offsets: Iterator,
        end_offsets: Iterator,
        binary_op: Op,
        h_init: Value,
        info: CommonData,
    ): ...
    def compute(
        self,
        temp_storage_ptr: int | None,
        temp_storage_nbytes: int,
        d_in: Iterator,
        d_out: Iterator,
        num_items: int,
        start_offsets: Iterator,
        end_offsets: Iterator,
        binary_op: Op,
        h_init: Value,
        max_segment_size: int | None = None,
        stream=None,
    ) -> int: ...

# ---------------
# DeviceMergeSort
# ---------------

class DeviceMergeSortBuildResult:
    def __init__(
        self,
        d_in_keys: Iterator,
        d_in_items: Iterator,
        d_out_keys: Iterator,
        d_out_items: Iterator,
        binary_op: Op,
        info: CommonData,
    ): ...
    def compute(
        self,
        temp_storage_ptr: int | None,
        temp_storage_nbytes: int,
        d_in_keys: Iterator,
        d_in_items: Iterator,
        d_out_keys: Iterator,
        d_out_items: Iterator,
        num_items: int,
        binary_op: Op,
        stream,
    ) -> int: ...

# -----------------
# DeviceUniqueByKey
# -----------------

class DeviceUniqueByKeyBuildResult:
    def __init__(
        self,
        d_keys_in: Iterator,
        d_values_in: Iterator,
        d_keys_out: Iterator,
        d_values_out: Iterator,
        d_num_selected_out: Iterator,
        binary_op: Op,
        info: CommonData,
    ): ...
    def compute(
        self,
        temp_storage_ptr: int | None,
        temp_storage_nbytes: int,
        d_keys_in: Iterator,
        d_values_in: Iterator,
        d_keys_out: Iterator,
        d_values_out: Iterator,
        d_num_selected_out: Iterator,
        binary_op: Op,
        num_items: int,
        stream,
    ) -> int: ...

# -----------------
# DeviceRadixSort
# -----------------

class DeviceRadixSortBuildResult:
    def __init__(self): ...
    def compute(
        self,
        temp_storage_ptr: int | None,
        temp_storage_nbytes: int,
        d_keys_in: Iterator,
        d_keys_out: Iterator,
        d_values_in: Iterator,
        d_values_out: Iterator,
        decomposer_op: Op,
        num_items: int,
        begin_bit: int,
        end_bit: int,
        is_overwrite_okay: bool,
        selector: int,
        stream,
    ) -> tuple[int, int]: ...

# --------------------
# DeviceUnaryTransform
# --------------------

class DeviceUnaryTransform:
    def __init__(
        self,
        d_in: Iterator,
        d_out: Iterator,
        op: Op,
        info: CommonData,
    ): ...
    def compute(
        self,
        d_in: Iterator,
        d_out: Iterator,
        num_items: int,
        stream,
    ) -> None: ...

# ---------------------
# DeviceBinaryTransform
# ---------------------

class DeviceBinaryTransform:
    def __init__(
        self,
        d_in1: Iterator,
        d_in2: Iterator,
        d_out: Iterator,
        op: Op,
        info: CommonData,
    ): ...
    def compute(
        self,
        d_in1: Iterator,
        d_in2: Iterator,
        d_out: Iterator,
        num_items: int,
        stream,
    ) -> None: ...

# ---------------
# DeviceHistogram
# ---------------

class DeviceHistogramBuildResult:
    def __init__(
        self,
        num_channels: int,
        num_active_channels: int,
        d_samples: Iterator,
        num_levels: int,
        d_histogram: Iterator,
        h_levels: Value,
        num_rows: int,
        row_stride_samples: int,
        is_evenly_segmented: bool,
        info: CommonData,
    ): ...
    def compute_even(
        self,
        d_samples: Iterator,
        d_histogram: Iterator,
        h_num_output_levels: Value,
        h_lower_level: Value,
        h_upper_level: Value,
        num_row_pixels: int,
        num_rows: int,
        row_stride_samples: int,
        stream,
    ) -> None: ...

# -------------------
# DeviceBinarySearch
# -------------------

class DeviceBinarySearchBuildResult:
    def __init__(
        self,
        mode: BinarySearchMode,
        d_data: Iterator,
        d_values: Iterator,
        d_out: Iterator,
        comparison_op: Op,
        info: CommonData,
    ): ...
    def compute(
        self,
        d_data: Iterator,
        num_items: int,
        d_values: Iterator,
        num_values: int,
        d_out: Iterator,
        comparison_op: Op,
        stream,
    ) -> None: ...

# -----------------
# DeviceSegmentedSort
# -----------------

class DeviceSegmentedSortBuildResult:
    def __init__(self): ...
    def compute(
        self,
        temp_storage_ptr: int | None,
        temp_storage_nbytes: int,
        d_in_keys: Iterator,
        d_out_keys: Iterator,
        d_in_values: Iterator,
        d_out_values: Iterator,
        num_items: int,
        num_segments: int,
        d_begin_offsets: Iterator,
        d_end_offsets: Iterator,
        is_overwrite_okay: bool,
        selector: int,
        stream,
    ) -> tuple[int, int]: ...

# ---------------------
# DeviceThreeWayPartition
# ---------------------

class DeviceThreeWayPartitionBuildResult:
    def __init__(
        self,
        d_in: Iterator,
        d_first_part_out: Iterator,
        d_second_part_out: Iterator,
        d_unselected_out: Iterator,
        d_num_selected_out: Iterator,
        select_first_part_op: Op,
        select_second_part_op: Op,
        info: CommonData,
    ): ...
    def compute(
        self,
        d_in: Iterator,
        d_first_part_out: Iterator,
        d_second_part_out: Iterator,
        d_unselected_out: Iterator,
        d_num_selected_out: Iterator,
        num_items: int,
        stream,
    ) -> int: ...



# Copyright (c) 2024-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception


from __future__ import annotations

from ... import _bindings, types
from ... import _cccl_interop as cccl
from ..._caching import cache_with_registered_key_functions
from ..._cccl_interop import call_build, set_cccl_iterator_state
from ..._utils.protocols import (
    get_data_pointer,
    validate_and_get_stream,
)
from ..._utils.temp_storage_buffer import TempStorageBuffer
from ...op import OpAdapter, make_op_adapter
from ...typing import DeviceArrayLike, IteratorT, Operator


class _MergeSort:
    __slots__ = [
        "d_in_keys_cccl",
        "d_in_items_cccl",
        "d_out_keys_cccl",
        "d_out_items_cccl",
        "op_adapter",
        "op_cccl",
        "build_result",
    ]

    def __init__(
        self,
        d_in_keys: DeviceArrayLike | IteratorT,
        d_in_items: DeviceArrayLike | IteratorT | None,
        d_out_keys: DeviceArrayLike,
        d_out_items: DeviceArrayLike | None,
        op: OpAdapter,
    ):
        present_in_values = d_in_items is not None
        present_out_values = d_out_items is not None
        assert present_in_values == present_out_values

        self.d_in_keys_cccl = cccl.to_cccl_input_iter(d_in_keys)
        self.d_in_items_cccl = cccl.to_cccl_input_iter(d_in_items)
        self.d_out_keys_cccl = cccl.to_cccl_output_iter(d_out_keys)
        self.d_out_items_cccl = cccl.to_cccl_output_iter(d_out_items)
        self.op_adapter = op

        # Compile the op - merge_sort expects int8 return (comparison)
        value_type = cccl.get_value_type(d_in_keys)
        self.op_cccl = op.compile((value_type, value_type), types.int8)

        self.build_result = call_build(
            _bindings.DeviceMergeSortBuildResult,
            self.d_in_keys_cccl,
            self.d_in_items_cccl,
            self.d_out_keys_cccl,
            self.d_out_items_cccl,
            self.op_cccl,
        )

    def __call__(
        self,
        temp_storage,
        d_in_keys: DeviceArrayLike | IteratorT,
        d_in_items: DeviceArrayLike | IteratorT | None,
        d_out_keys: DeviceArrayLike,
        d_out_items: DeviceArrayLike | None,
        op: Operator,
        num_items: int,
        stream=None,
    ):
        present_in_values = d_in_items is not None
        present_out_values = d_out_items is not None
        assert present_in_values == present_out_values

        set_cccl_iterator_state(self.d_in_keys_cccl, d_in_keys)
        if present_in_values:
            set_cccl_iterator_state(self.d_in_items_cccl, d_in_items)
        set_cccl_iterator_state(self.d_out_keys_cccl, d_out_keys)
        if present_out_values:
            set_cccl_iterator_state(self.d_out_items_cccl, d_out_items)

        op_adapter = make_op_adapter(op)
        self.op_cccl.state = op_adapter.get_state()

        stream_handle = validate_and_get_stream(stream)
        if temp_storage is None:
            temp_storage_bytes = 0
            d_temp_storage = 0
        else:
            temp_storage_bytes = temp_storage.nbytes
            # Note: this is slightly slower, but supports all ndarray-like objects as long as they support CAI
            # TODO: switch to use gpumemoryview once it's ready
            d_temp_storage = get_data_pointer(temp_storage)

        temp_storage_bytes = self.build_result.compute(
            d_temp_storage,
            temp_storage_bytes,
            self.d_in_keys_cccl,
            self.d_in_items_cccl,
            self.d_out_keys_cccl,
            self.d_out_items_cccl,
            num_items,
            self.op_cccl,
            stream_handle,
        )

        return temp_storage_bytes


@cache_with_registered_key_functions
def make_merge_sort(
    d_in_keys: DeviceArrayLike | IteratorT,
    d_in_items: DeviceArrayLike | IteratorT | None,
    d_out_keys: DeviceArrayLike,
    d_out_items: DeviceArrayLike | None,
    op: Operator,
):
    """Implements a device-wide merge sort using ``d_in_keys`` and the comparison operator ``op``.

    Example:
        Below, ``make_merge_sort`` is used to create a merge sort object that can be reused.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/sort/merge_sort_object.py
          :language: python
          :start-after: # example-begin


    Args:
        d_in_keys: Device array or iterator containing the input keys to be sorted
        d_in_items: Optional device array or iterator that contains each key's corresponding item
        d_out_keys: Device array to store the sorted keys
        d_out_items: Device array to store the sorted items
        op: The comparison operator for sorting. The signature is  ``(T, T) -> int8``, where ``T`` is the input data type. See notes below.

    Returns:
        A callable object that can be used to perform the merge sort

    .. important::

      The provided comparison operator must follow `strict weak ordering <https://en.cppreference.com/w/cpp/concepts/strict_weak_order.html>`_
      semantics. For example, the comparator ``lambda lhs, rhs: lhs < rhs``  follows strict weak ordering,  but the comparator
      ``lambda lhs, rhs: rhs >= lhs`` does not, because it is reflexive: ``r(x, x) == True``. Providing a comparator that does not
      follow the required semantics can lead to incorrect results, silent memory corruption, or crashes.
    """
    op_adapter = make_op_adapter(op)
    return _MergeSort(d_in_keys, d_in_items, d_out_keys, d_out_items, op_adapter)


def merge_sort(
    d_in_keys: DeviceArrayLike | IteratorT,
    d_in_items: DeviceArrayLike | IteratorT | None,
    d_out_keys: DeviceArrayLike,
    d_out_items: DeviceArrayLike | None,
    op: Operator,
    num_items: int,
    stream=None,
):
    """
    Performs device-wide merge sort.

    This function automatically handles temporary storage allocation and execution.

    Example:
        Below, ``merge_sort`` is used to sort a sequence of keys inplace. It also rearranges the items according to the keys' order.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/sort/merge_sort_basic.py
            :language: python
            :start-after: # example-begin

    Args:
        d_in_keys: Device array or iterator containing the input sequence of keys
        d_in_items: Device array or iterator containing the input sequence of items (optional)
        d_out_keys: Device array to store the sorted keys
        d_out_items: Device array to store the sorted items (optional)
        op: The comparison operator for sorting. The signature is  ``(T, T) -> int8``, where ``T`` is the input data type. See notes below.
        num_items: Number of items to sort
        stream: CUDA stream for the operation (optional)

    .. important::

      The provided comparison operator must follow `strict weak ordering <https://en.cppreference.com/w/cpp/concepts/strict_weak_order.html>`_
      semantics. For example, the comparator ``lambda lhs, rhs: lhs < rhs``  follows strict weak ordering,  but the comparator
      ``lambda lhs, rhs: rhs >= lhs`` does not, because it is reflexive: ``r(x, x) == True``. Providing a comparator that does not
      follow the required semantics can lead to incorrect results, silent memory corruption, or crashes.
    """
    sorter = make_merge_sort(d_in_keys, d_in_items, d_out_keys, d_out_items, op)
    tmp_storage_bytes = sorter(
        None, d_in_keys, d_in_items, d_out_keys, d_out_items, op, num_items, stream
    )
    tmp_storage = TempStorageBuffer(tmp_storage_bytes, stream)
    sorter(
        tmp_storage,
        d_in_keys,
        d_in_items,
        d_out_keys,
        d_out_items,
        op,
        num_items,
        stream,
    )
